import pytest


@pytest.fixture
def sample_document_and_expected_formats():
    document = {
        "title": "The Birth of Machine Experience Engineering",
        "documentId": "1234567890",
        "body": {
            "content": [
                {
                    "endIndex": 1,
                    "sectionBreak": {
                        "sectionStyle": {
                            "columnSeparatorStyle": "NONE",
                            "contentDirection": "LEFT_TO_RIGHT",
                            "sectionType": "CONTINUOUS",
                        }
                    },
                },
                {
                    "startIndex": 1,
                    "endIndex": 45,
                    "paragraph": {
                        "elements": [
                            {
                                "endIndex": 45,
                                "startIndex": 1,
                                "textRun": {
                                    "content": "The Birth of Machine Experience Engineering\n",
                                    "textStyle": {
                                        "bold": True,
                                        "fontSize": {"magnitude": 23, "unit": "PT"},
                                    },
                                },
                            }
                        ],
                        "paragraphStyle": {
                            "direction": "LEFT_TO_RIGHT",
                            "headingId": "h.wwd7ec37bh6k",
                            "keepLinesTogether": False,
                            "keepWithNext": False,
                            "namedStyleType": "HEADING_1",
                            "spaceAbove": {"magnitude": 24, "unit": "PT"},
                        },
                    },
                },
                {
                    "startIndex": 45,
                    "endIndex": 46,
                    "paragraph": {
                        "elements": [
                            {
                                "startIndex": 304,
                                "endIndex": 305,
                                "inlineObjectElement": {
                                    "inlineObjectId": "kix.2s5wy5oiaf79",
                                    "textStyle": {},
                                },
                            },
                            {
                                "endIndex": 46,
                                "startIndex": 45,
                                "textRun": {"content": "\n", "textStyle": {}},
                            },
                        ],
                        "paragraphStyle": {
                            "direction": "LEFT_TO_RIGHT",
                            "namedStyleType": "NORMAL_TEXT",
                            "spaceAbove": {"magnitude": 12, "unit": "PT"},
                            "spaceBelow": {"magnitude": 12, "unit": "PT"},
                        },
                    },
                },
                {
                    "startIndex": 46,
                    "endIndex": 297,
                    "paragraph": {
                        "elements": [
                            {
                                "startIndex": 46,
                                "endIndex": 146,
                                "textRun": {
                                    "content": (
                                        "LLMs acting on behalf of humans and interacting with real-"
                                        "world systems isn't theoretical anymore - "
                                    ),
                                    "textStyle": {},
                                },
                            },
                            {
                                "startIndex": 146,
                                "endIndex": 175,
                                "textRun": {
                                    "content": "Arcade has made it a reality.",
                                    "textStyle": {
                                        "bold": True,
                                        "italic": True,
                                    },
                                },
                            },
                            {
                                "startIndex": 175,
                                "endIndex": 248,
                                "textRun": {
                                    "content": (
                                        " With this shift, we're seeing the emergence of a new "
                                        "software practice: "
                                    ),
                                    "textStyle": {},
                                },
                            },
                            {
                                "startIndex": 248,
                                "endIndex": 295,
                                "textRun": {
                                    "content": "Machine Experience Engineering (MX Engineering)",
                                    "textStyle": {
                                        "italic": True,
                                    },
                                },
                            },
                            {
                                "startIndex": 295,
                                "endIndex": 297,
                                "textRun": {
                                    "content": ".\n",
                                    "textStyle": {},
                                },
                            },
                        ],
                        "paragraphStyle": {
                            "direction": "LEFT_TO_RIGHT",
                            "namedStyleType": "NORMAL_TEXT",
                            "spaceAbove": {"magnitude": 12, "unit": "PT"},
                            "spaceBelow": {"magnitude": 12, "unit": "PT"},
                        },
                    },
                },
                {
                    "endIndex": 407,
                    "startIndex": 297,
                    "table": {
                        "columns": 3,
                        "rows": 3,
                        "tableRows": [
                            {
                                "endIndex": 338,
                                "startIndex": 297,
                                "tableCells": [
                                    {
                                        "content": [
                                            {
                                                "endIndex": 318,
                                                "paragraph": {
                                                    "elements": [
                                                        {
                                                            "endIndex": 318,
                                                            "startIndex": 309,
                                                            "textRun": {
                                                                "content": "Column 1\n",
                                                                "textStyle": {"bold": True},
                                                            },
                                                        }
                                                    ],
                                                    "paragraphStyle": {
                                                        "alignment": "START",
                                                        "avoidWidowAndOrphan": False,
                                                        "borderBetween": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderBottom": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderLeft": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderRight": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderTop": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "direction": "LEFT_TO_RIGHT",
                                                        "indentEnd": {"unit": "PT"},
                                                        "indentFirstLine": {"unit": "PT"},
                                                        "indentStart": {"unit": "PT"},
                                                        "keepLinesTogether": False,
                                                        "keepWithNext": False,
                                                        "lineSpacing": 100,
                                                        "namedStyleType": "NORMAL_TEXT",
                                                        "pageBreakBefore": False,
                                                        "shading": {"backgroundColor": {}},
                                                        "spaceAbove": {"unit": "PT"},
                                                        "spaceBelow": {"unit": "PT"},
                                                        "spacingMode": "COLLAPSE_LISTS",
                                                    },
                                                },
                                                "startIndex": 309,
                                            }
                                        ],
                                        "endIndex": 318,
                                        "startIndex": 308,
                                        "tableCellStyle": {
                                            "backgroundColor": {},
                                            "columnSpan": 1,
                                            "contentAlignment": "TOP",
                                            "paddingBottom": {"magnitude": 5, "unit": "PT"},
                                            "paddingLeft": {"magnitude": 5, "unit": "PT"},
                                            "paddingRight": {"magnitude": 5, "unit": "PT"},
                                            "paddingTop": {"magnitude": 5, "unit": "PT"},
                                            "rowSpan": 1,
                                        },
                                    },
                                    {
                                        "content": [
                                            {
                                                "endIndex": 334,
                                                "paragraph": {
                                                    "elements": [
                                                        {
                                                            "endIndex": 326,
                                                            "startIndex": 319,
                                                            "textRun": {
                                                                "content": "Another",
                                                                "textStyle": {"italic": True},
                                                            },
                                                        },
                                                        {
                                                            "endIndex": 334,
                                                            "startIndex": 326,
                                                            "textRun": {
                                                                "content": " column\n",
                                                                "textStyle": {},
                                                            },
                                                        },
                                                    ],
                                                    "paragraphStyle": {
                                                        "alignment": "START",
                                                        "avoidWidowAndOrphan": False,
                                                        "borderBetween": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderBottom": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderLeft": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderRight": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderTop": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "direction": "LEFT_TO_RIGHT",
                                                        "indentEnd": {"unit": "PT"},
                                                        "indentFirstLine": {"unit": "PT"},
                                                        "indentStart": {"unit": "PT"},
                                                        "keepLinesTogether": False,
                                                        "keepWithNext": False,
                                                        "lineSpacing": 100,
                                                        "namedStyleType": "NORMAL_TEXT",
                                                        "pageBreakBefore": False,
                                                        "shading": {"backgroundColor": {}},
                                                        "spaceAbove": {"unit": "PT"},
                                                        "spaceBelow": {"unit": "PT"},
                                                        "spacingMode": "COLLAPSE_LISTS",
                                                    },
                                                },
                                                "startIndex": 319,
                                            }
                                        ],
                                        "endIndex": 334,
                                        "startIndex": 318,
                                        "tableCellStyle": {
                                            "backgroundColor": {},
                                            "columnSpan": 1,
                                            "contentAlignment": "TOP",
                                            "paddingBottom": {"magnitude": 5, "unit": "PT"},
                                            "paddingLeft": {"magnitude": 5, "unit": "PT"},
                                            "paddingRight": {"magnitude": 5, "unit": "PT"},
                                            "paddingTop": {"magnitude": 5, "unit": "PT"},
                                            "rowSpan": 1,
                                        },
                                    },
                                    {
                                        "content": [
                                            {
                                                "endIndex": 348,
                                                "paragraph": {
                                                    "elements": [
                                                        {
                                                            "endIndex": 348,
                                                            "startIndex": 335,
                                                            "textRun": {
                                                                "content": "Third column\n",
                                                                "textStyle": {},
                                                            },
                                                        }
                                                    ],
                                                    "paragraphStyle": {
                                                        "alignment": "START",
                                                        "avoidWidowAndOrphan": False,
                                                        "borderBetween": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderBottom": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderLeft": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderRight": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderTop": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "direction": "LEFT_TO_RIGHT",
                                                        "indentEnd": {"unit": "PT"},
                                                        "indentFirstLine": {"unit": "PT"},
                                                        "indentStart": {"unit": "PT"},
                                                        "keepLinesTogether": False,
                                                        "keepWithNext": False,
                                                        "lineSpacing": 100,
                                                        "namedStyleType": "NORMAL_TEXT",
                                                        "pageBreakBefore": False,
                                                        "shading": {"backgroundColor": {}},
                                                        "spaceAbove": {"unit": "PT"},
                                                        "spaceBelow": {"unit": "PT"},
                                                        "spacingMode": "COLLAPSE_LISTS",
                                                    },
                                                },
                                                "startIndex": 335,
                                            }
                                        ],
                                        "endIndex": 348,
                                        "startIndex": 334,
                                        "tableCellStyle": {
                                            "backgroundColor": {},
                                            "columnSpan": 1,
                                            "contentAlignment": "TOP",
                                            "paddingBottom": {"magnitude": 5, "unit": "PT"},
                                            "paddingLeft": {"magnitude": 5, "unit": "PT"},
                                            "paddingRight": {"magnitude": 5, "unit": "PT"},
                                            "paddingTop": {"magnitude": 5, "unit": "PT"},
                                            "rowSpan": 1,
                                        },
                                    },
                                ],
                                "tableRowStyle": {"minRowHeight": {"unit": "PT"}},
                            },
                            {
                                "endIndex": 366,
                                "startIndex": 348,
                                "tableCells": [
                                    {
                                        "content": [
                                            {
                                                "endIndex": 356,
                                                "paragraph": {
                                                    "elements": [
                                                        {
                                                            "endIndex": 356,
                                                            "startIndex": 350,
                                                            "textRun": {
                                                                "content": "Hello\n",
                                                                "textStyle": {},
                                                            },
                                                        }
                                                    ],
                                                    "paragraphStyle": {
                                                        "alignment": "START",
                                                        "avoidWidowAndOrphan": False,
                                                        "borderBetween": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderBottom": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderLeft": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderRight": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderTop": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "direction": "LEFT_TO_RIGHT",
                                                        "indentEnd": {"unit": "PT"},
                                                        "indentFirstLine": {"unit": "PT"},
                                                        "indentStart": {"unit": "PT"},
                                                        "keepLinesTogether": False,
                                                        "keepWithNext": False,
                                                        "lineSpacing": 100,
                                                        "namedStyleType": "NORMAL_TEXT",
                                                        "pageBreakBefore": False,
                                                        "shading": {"backgroundColor": {}},
                                                        "spaceAbove": {"unit": "PT"},
                                                        "spaceBelow": {"unit": "PT"},
                                                        "spacingMode": "COLLAPSE_LISTS",
                                                    },
                                                },
                                                "startIndex": 350,
                                            }
                                        ],
                                        "endIndex": 356,
                                        "startIndex": 349,
                                        "tableCellStyle": {
                                            "backgroundColor": {},
                                            "columnSpan": 1,
                                            "contentAlignment": "TOP",
                                            "paddingBottom": {"magnitude": 5, "unit": "PT"},
                                            "paddingLeft": {"magnitude": 5, "unit": "PT"},
                                            "paddingRight": {"magnitude": 5, "unit": "PT"},
                                            "paddingTop": {"magnitude": 5, "unit": "PT"},
                                            "rowSpan": 1,
                                        },
                                    },
                                    {
                                        "content": [
                                            {
                                                "endIndex": 364,
                                                "paragraph": {
                                                    "elements": [
                                                        {
                                                            "endIndex": 364,
                                                            "startIndex": 357,
                                                            "textRun": {
                                                                "content": "world!\n",
                                                                "textStyle": {},
                                                            },
                                                        }
                                                    ],
                                                    "paragraphStyle": {
                                                        "alignment": "START",
                                                        "avoidWidowAndOrphan": False,
                                                        "borderBetween": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderBottom": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderLeft": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderRight": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderTop": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "direction": "LEFT_TO_RIGHT",
                                                        "indentEnd": {"unit": "PT"},
                                                        "indentFirstLine": {"unit": "PT"},
                                                        "indentStart": {"unit": "PT"},
                                                        "keepLinesTogether": False,
                                                        "keepWithNext": False,
                                                        "lineSpacing": 100,
                                                        "namedStyleType": "NORMAL_TEXT",
                                                        "pageBreakBefore": False,
                                                        "shading": {"backgroundColor": {}},
                                                        "spaceAbove": {"unit": "PT"},
                                                        "spaceBelow": {"unit": "PT"},
                                                        "spacingMode": "COLLAPSE_LISTS",
                                                    },
                                                },
                                                "startIndex": 357,
                                            }
                                        ],
                                        "endIndex": 364,
                                        "startIndex": 356,
                                        "tableCellStyle": {
                                            "backgroundColor": {},
                                            "columnSpan": 1,
                                            "contentAlignment": "TOP",
                                            "paddingBottom": {"magnitude": 5, "unit": "PT"},
                                            "paddingLeft": {"magnitude": 5, "unit": "PT"},
                                            "paddingRight": {"magnitude": 5, "unit": "PT"},
                                            "paddingTop": {"magnitude": 5, "unit": "PT"},
                                            "rowSpan": 1,
                                        },
                                    },
                                    {
                                        "content": [
                                            {
                                                "endIndex": 366,
                                                "paragraph": {
                                                    "elements": [
                                                        {
                                                            "endIndex": 366,
                                                            "startIndex": 365,
                                                            "textRun": {
                                                                "content": "\n",
                                                                "textStyle": {},
                                                            },
                                                        }
                                                    ],
                                                    "paragraphStyle": {
                                                        "alignment": "START",
                                                        "avoidWidowAndOrphan": False,
                                                        "borderBetween": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderBottom": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderLeft": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderRight": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderTop": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "direction": "LEFT_TO_RIGHT",
                                                        "indentEnd": {"unit": "PT"},
                                                        "indentFirstLine": {"unit": "PT"},
                                                        "indentStart": {"unit": "PT"},
                                                        "keepLinesTogether": False,
                                                        "keepWithNext": False,
                                                        "lineSpacing": 100,
                                                        "namedStyleType": "NORMAL_TEXT",
                                                        "pageBreakBefore": False,
                                                        "shading": {"backgroundColor": {}},
                                                        "spaceAbove": {"unit": "PT"},
                                                        "spaceBelow": {"unit": "PT"},
                                                        "spacingMode": "COLLAPSE_LISTS",
                                                    },
                                                },
                                                "startIndex": 365,
                                            }
                                        ],
                                        "endIndex": 366,
                                        "startIndex": 364,
                                        "tableCellStyle": {
                                            "backgroundColor": {},
                                            "columnSpan": 1,
                                            "contentAlignment": "TOP",
                                            "paddingBottom": {"magnitude": 5, "unit": "PT"},
                                            "paddingLeft": {"magnitude": 5, "unit": "PT"},
                                            "paddingRight": {"magnitude": 5, "unit": "PT"},
                                            "paddingTop": {"magnitude": 5, "unit": "PT"},
                                            "rowSpan": 1,
                                        },
                                    },
                                ],
                                "tableRowStyle": {"minRowHeight": {"unit": "PT"}},
                            },
                            {
                                "endIndex": 415,
                                "startIndex": 366,
                                "tableCells": [
                                    {
                                        "content": [
                                            {
                                                "endIndex": 388,
                                                "paragraph": {
                                                    "elements": [
                                                        {
                                                            "endIndex": 388,
                                                            "startIndex": 368,
                                                            "textRun": {
                                                                "content": "The quick brown fox\n",
                                                                "textStyle": {},
                                                            },
                                                        }
                                                    ],
                                                    "paragraphStyle": {
                                                        "alignment": "START",
                                                        "avoidWidowAndOrphan": False,
                                                        "borderBetween": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderBottom": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderLeft": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderRight": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderTop": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "direction": "LEFT_TO_RIGHT",
                                                        "indentEnd": {"unit": "PT"},
                                                        "indentFirstLine": {"unit": "PT"},
                                                        "indentStart": {"unit": "PT"},
                                                        "keepLinesTogether": False,
                                                        "keepWithNext": False,
                                                        "lineSpacing": 100,
                                                        "namedStyleType": "NORMAL_TEXT",
                                                        "pageBreakBefore": False,
                                                        "shading": {"backgroundColor": {}},
                                                        "spaceAbove": {"unit": "PT"},
                                                        "spaceBelow": {"unit": "PT"},
                                                        "spacingMode": "COLLAPSE_LISTS",
                                                    },
                                                },
                                                "startIndex": 368,
                                            }
                                        ],
                                        "endIndex": 388,
                                        "startIndex": 367,
                                        "tableCellStyle": {
                                            "backgroundColor": {},
                                            "columnSpan": 1,
                                            "contentAlignment": "TOP",
                                            "paddingBottom": {"magnitude": 5, "unit": "PT"},
                                            "paddingLeft": {"magnitude": 5, "unit": "PT"},
                                            "paddingRight": {"magnitude": 5, "unit": "PT"},
                                            "paddingTop": {"magnitude": 5, "unit": "PT"},
                                            "rowSpan": 1,
                                        },
                                    },
                                    {
                                        "content": [
                                            {
                                                "endIndex": 401,
                                                "paragraph": {
                                                    "elements": [
                                                        {
                                                            "endIndex": 395,
                                                            "startIndex": 389,
                                                            "textRun": {
                                                                "content": "jumped",
                                                                "textStyle": {"italic": True},
                                                            },
                                                        },
                                                        {
                                                            "endIndex": 401,
                                                            "startIndex": 395,
                                                            "textRun": {
                                                                "content": " over\n",
                                                                "textStyle": {},
                                                            },
                                                        },
                                                    ],
                                                    "paragraphStyle": {
                                                        "alignment": "START",
                                                        "avoidWidowAndOrphan": False,
                                                        "borderBetween": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderBottom": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderLeft": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderRight": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderTop": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "direction": "LEFT_TO_RIGHT",
                                                        "indentEnd": {"unit": "PT"},
                                                        "indentFirstLine": {"unit": "PT"},
                                                        "indentStart": {"unit": "PT"},
                                                        "keepLinesTogether": False,
                                                        "keepWithNext": False,
                                                        "lineSpacing": 100,
                                                        "namedStyleType": "NORMAL_TEXT",
                                                        "pageBreakBefore": False,
                                                        "shading": {"backgroundColor": {}},
                                                        "spaceAbove": {"unit": "PT"},
                                                        "spaceBelow": {"unit": "PT"},
                                                        "spacingMode": "COLLAPSE_LISTS",
                                                    },
                                                },
                                                "startIndex": 389,
                                            }
                                        ],
                                        "endIndex": 401,
                                        "startIndex": 388,
                                        "tableCellStyle": {
                                            "backgroundColor": {},
                                            "columnSpan": 1,
                                            "contentAlignment": "TOP",
                                            "paddingBottom": {"magnitude": 5, "unit": "PT"},
                                            "paddingLeft": {"magnitude": 5, "unit": "PT"},
                                            "paddingRight": {"magnitude": 5, "unit": "PT"},
                                            "paddingTop": {"magnitude": 5, "unit": "PT"},
                                            "rowSpan": 1,
                                        },
                                    },
                                    {
                                        "content": [
                                            {
                                                "endIndex": 415,
                                                "paragraph": {
                                                    "elements": [
                                                        {
                                                            "endIndex": 415,
                                                            "startIndex": 402,
                                                            "textRun": {
                                                                "content": "the lazy dog\n",
                                                                "textStyle": {},
                                                            },
                                                        }
                                                    ],
                                                    "paragraphStyle": {
                                                        "alignment": "START",
                                                        "avoidWidowAndOrphan": False,
                                                        "borderBetween": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderBottom": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderLeft": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderRight": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "borderTop": {
                                                            "color": {},
                                                            "dashStyle": "SOLID",
                                                            "padding": {"unit": "PT"},
                                                            "width": {"unit": "PT"},
                                                        },
                                                        "direction": "LEFT_TO_RIGHT",
                                                        "indentEnd": {"unit": "PT"},
                                                        "indentFirstLine": {"unit": "PT"},
                                                        "indentStart": {"unit": "PT"},
                                                        "keepLinesTogether": False,
                                                        "keepWithNext": False,
                                                        "lineSpacing": 100,
                                                        "namedStyleType": "NORMAL_TEXT",
                                                        "pageBreakBefore": False,
                                                        "shading": {"backgroundColor": {}},
                                                        "spaceAbove": {"unit": "PT"},
                                                        "spaceBelow": {"unit": "PT"},
                                                        "spacingMode": "COLLAPSE_LISTS",
                                                    },
                                                },
                                                "startIndex": 402,
                                            }
                                        ],
                                        "endIndex": 415,
                                        "startIndex": 401,
                                        "tableCellStyle": {
                                            "backgroundColor": {},
                                            "columnSpan": 1,
                                            "contentAlignment": "TOP",
                                            "paddingBottom": {"magnitude": 5, "unit": "PT"},
                                            "paddingLeft": {"magnitude": 5, "unit": "PT"},
                                            "paddingRight": {"magnitude": 5, "unit": "PT"},
                                            "paddingTop": {"magnitude": 5, "unit": "PT"},
                                            "rowSpan": 1,
                                        },
                                    },
                                ],
                                "tableRowStyle": {"minRowHeight": {"unit": "PT"}},
                            },
                        ],
                        "tableStyle": {
                            "tableColumnProperties": [
                                {"widthType": "EVENLY_DISTRIBUTED"},
                                {"widthType": "EVENLY_DISTRIBUTED"},
                                {"widthType": "EVENLY_DISTRIBUTED"},
                            ]
                        },
                    },
                },
            ]
        },
    }

    expected_markdown = (
        "---\ntitle: The Birth of Machine Experience Engineering\ndocumentId: 1234567890\n---\n"
        "# **The Birth of Machine Experience Engineering**\n"
        "\n"
        "LLMs acting on behalf of humans and interacting with real-world systems isn't theoretical "
        "anymore - "
        "**_Arcade has made it a reality._** With this shift, we're seeing the emergence of a new "
        "software practice: "
        "_Machine Experience Engineering (MX Engineering)_.\n"
        "<table>"
        "<tr>"
        "<td><b>Column 1</b></td>"
        "<td><i>Another</i> column</td>"
        "<td>Third column</td>"
        "</tr>"
        "<tr>"
        "<td>Hello</td>"
        "<td>world!</td>"
        "<td></td>"
        "</tr>"
        "<tr>"
        "<td>The quick brown fox</td>"
        "<td><i>jumped</i> over</td>"
        "<td>the lazy dog</td>"
        "</tr>"
        "</table>"
    )

    expected_html = (
        "<html><head>"
        "<title>The Birth of Machine Experience Engineering</title>"
        '<meta name="documentId" content="1234567890">'
        "</head><body>"
        "<h1><b>The Birth of Machine Experience Engineering</b></h1>"
        "<p>LLMs acting on behalf of humans and interacting with real-world systems isn't "
        "theoretical anymore - "
        "<b><i>Arcade has made it a reality.</i></b> With this shift, we're seeing the emergence "
        "of a new software practice: <i>Machine Experience Engineering (MX Engineering)</i>.</p>"
        "<table>"
        "<tr>"
        "<td><b>Column 1</b></td>"
        "<td><i>Another</i> column</td>"
        "<td>Third column</td>"
        "</tr>"
        "<tr>"
        "<td>Hello</td>"
        "<td>world!</td>"
        "<td></td>"
        "</tr>"
        "<tr>"
        "<td>The quick brown fox</td>"
        "<td><i>jumped</i> over</td>"
        "<td>the lazy dog</td>"
        "</tr>"
        "</table>"
        "</body></html>"
    )

    return document, expected_markdown, expected_html



from unittest.mock import MagicMock

import httpx
import pytest

from arcade_dropbox.constants import ItemCategory
from arcade_dropbox.tools.browse import search_files_and_folders
from arcade_dropbox.utils import clean_dropbox_entry


@pytest.fixture
def sample_folder_match(sample_folder_entry):
    return {
        "metadata": {
            ".tag": "metadata",
            "metadata": sample_folder_entry,
        }
    }


@pytest.fixture
def sample_file_match(sample_file_entry):
    return {
        "metadata": {
            ".tag": "metadata",
            "metadata": sample_file_entry,
        }
    }


@pytest.mark.asyncio
async def test_search_files_success_empty_results(
    mock_context,
    mock_httpx_client,
):
    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {"matches": [], "cursor": None, "has_more": False}
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await search_files_and_folders(
        context=mock_context,
        keywords="do not match anything",
    )

    assert tool_response == {
        "items": [],
        "cursor": None,
        "has_more": False,
    }


@pytest.mark.asyncio
async def test_search_files_success_with_matches(
    mock_context,
    mock_httpx_client,
    sample_file_match,
    sample_folder_match,
    sample_file_entry,
    sample_folder_entry,
):
    matches = [
        sample_file_match,
        sample_folder_match,
    ]

    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {"matches": matches, "cursor": None, "has_more": False}
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await search_files_and_folders(
        context=mock_context,
        keywords="test",
    )

    assert tool_response == {
        "items": [
            clean_dropbox_entry(sample_file_entry),
            clean_dropbox_entry(sample_folder_entry),
        ],
        "cursor": None,
        "has_more": False,
    }


@pytest.mark.asyncio
async def test_search_files_success_with_path_missing_leading_slash(
    mock_context,
    mock_httpx_client,
    sample_file_match,
    sample_folder_match,
    sample_file_entry,
    sample_folder_entry,
):
    matches = [
        sample_file_match,
        sample_folder_match,
    ]

    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {"matches": matches, "cursor": None, "has_more": False}
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await search_files_and_folders(
        context=mock_context,
        keywords="test",
        search_in_folder_path="TestFolder",
    )

    assert tool_response == {
        "items": [
            clean_dropbox_entry(sample_file_entry),
            clean_dropbox_entry(sample_folder_entry),
        ],
        "cursor": None,
        "has_more": False,
    }

    mock_httpx_client.post.assert_called_once_with(
        url="https://api.dropboxapi.com/2/files/search_v2",
        headers={"Authorization": "Bearer fake-token"},
        json={
            "query": "test",
            "options": {
                "file_categories": [],
                "path": "/TestFolder",
                "file_status": "active",
                "filename_only": False,
                "max_results": 100,
            },
        },
    )


@pytest.mark.asyncio
async def test_search_files_success_with_more_results_to_paginate(
    mock_context,
    mock_httpx_client,
    sample_file_match,
    sample_folder_match,
    sample_file_entry,
    sample_folder_entry,
):
    matches = [
        sample_file_match,
        sample_folder_match,
    ]

    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {
        "matches": matches,
        "cursor": "cursor",
        "has_more": True,
    }
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await search_files_and_folders(
        context=mock_context,
        keywords="test",
    )

    assert tool_response == {
        "items": [
            clean_dropbox_entry(sample_file_entry),
            clean_dropbox_entry(sample_folder_entry),
        ],
        "cursor": "cursor",
        "has_more": True,
    }


@pytest.mark.asyncio
async def test_search_files_success_providing_pagination_cursor(
    mock_context,
    mock_httpx_client,
    sample_file_match,
    sample_folder_match,
    sample_file_entry,
    sample_folder_entry,
):
    matches = [
        sample_file_match,
        sample_folder_match,
    ]

    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {"matches": matches, "cursor": None, "has_more": False}
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await search_files_and_folders(
        context=mock_context,
        keywords="test",
        cursor="cursor",
    )

    assert tool_response == {
        "items": [
            clean_dropbox_entry(sample_file_entry),
            clean_dropbox_entry(sample_folder_entry),
        ],
        "cursor": None,
        "has_more": False,
    }

    # Assert that the request was made with the correct cursor and not other arguments
    mock_httpx_client.post.assert_called_once_with(
        url="https://api.dropboxapi.com/2/files/search_v2/continue",
        headers={"Authorization": "Bearer fake-token"},
        json={"cursor": "cursor"},
    )


@pytest.mark.asyncio
async def test_search_files_path_not_found(
    mock_context,
    mock_httpx_client,
):
    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 409
    mock_httpx_response.json.return_value = {"error_summary": "path/not_found"}

    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await search_files_and_folders(
        context=mock_context,
        keywords="test",
        search_in_folder_path="/not/exist/folder",
    )

    assert tool_response == {
        "error": "The specified path was not found by Dropbox",
    }


@pytest.mark.asyncio
async def test_search_files_success_filtering_by_category(
    mock_context,
    mock_httpx_client,
    sample_file_match,
    sample_folder_match,
    sample_file_entry,
    sample_folder_entry,
):
    matches = [
        sample_file_match,
        sample_folder_match,
    ]

    mock_httpx_response = MagicMock(spec=httpx.Response)
    mock_httpx_response.status_code = 200
    mock_httpx_response.json.return_value = {"matches": matches, "cursor": None, "has_more": False}
    mock_httpx_client.post.return_value = mock_httpx_response

    tool_response = await search_files_and_folders(
        context=mock_context,
        keywords="test",
        filter_by_category=[ItemCategory.PDF],
    )

    assert tool_response == {
        "items": [
            clean_dropbox_entry(sample_file_entry),
            clean_dropbox_entry(sample_folder_entry),
        ],
        "cursor": None,
        "has_more": False,
    }

    mock_httpx_client.post.assert_called_once_with(
        url="https://api.dropboxapi.com/2/files/search_v2",
        headers={"Authorization": "Bearer fake-token"},
        json={
            "query": "test",
            "options": {
                "path": "",
                "file_status": "active",
                "filename_only": False,
                "max_results": 100,
                "file_categories": [ItemCategory.PDF.value],
            },
        },
    )



from unittest.mock import MagicMock, patch

import pytest
from arcade_tdk.errors import RetryableToolError, ToolExecutionError

from arcade_google.models import (
    CellData,
    CellExtendedValue,
    NumberFormatType,
    RowData,
    SheetDataInput,
)
from arcade_google.utils import (
    col_to_index,
    compute_sheet_data_dimensions,
    convert_api_grid_data_to_dict,
    create_cell_data,
    create_row_data,
    create_sheet_data,
    create_sheet_properties,
    extract_user_entered_cell_value,
    group_contiguous_rows,
    index_to_col,
    is_col_greater,
    process_row,
    validate_write_to_cell_params,
)


@pytest.fixture
def sheet_data_input_fixture():
    data = {
        1: {
            "A": "name",
            "B": "age",
            "C": "email",
            "D": "score",
            "E": "gender",
            "F": "city",
            "G": "country",
            "H": "registration_date",
        },
        2: {
            "A": "John Doe",
            "B": 28,
            "C": "johndoe@example.com",
            "D": 85.4,
            "E": "Male",
            "F": "New York",
            "G": "USA",
            "H": "2023-01-15",
        },
        10: {
            "A": "Nate Green",
            "B": 30,
            "C": "nateg@example.com",
            "D": 88,
            "E": "Male",
            "F": "Orlando",
            "G": "USA",
            "H": "2024-02-01",
        },
        43: {
            "A": 100,
            "B": 300,
            "H": 123,
            "I": "=SUM(SEQUENCE(10))",
        },
        44: {
            "A": 456,
            "B": 234,
            "H": 123,
            "I": "=SUM(SEQUENCE(10))",
        },
    }
    return SheetDataInput(data=data)


@pytest.mark.parametrize(
    "col, expected_index",
    [
        ("A", 0),
        ("B", 1),
        ("Z", 25),
        ("AA", 26 + 0),
        ("AZ", (1 * 26) + 25),
        ("BA", (2 * 26) + 0),
        ("ZZ", (26 * 26) + 25),
        ("AAA", (1 * 26 * 26) + (1 * 26) + 0),
        ("AAB", (1 * 26 * 26) + (1 * 26) + 1),
        ("QED", (17 * 26 * 26) + (5 * 26) + 3),
    ],
)
def test_col_to_index(col, expected_index):
    assert col_to_index(col) == expected_index


@pytest.mark.parametrize(
    "index, expected_col",
    [
        (0, "A"),
        (1, "B"),
        (25, "Z"),
        (26 + 0, "AA"),
        ((1 * 26) + 25, "AZ"),
        ((2 * 26) + 0, "BA"),
        ((26 * 26) + 25, "ZZ"),
        ((1 * 26 * 26) + (1 * 26) + 0, "AAA"),
        ((1 * 26 * 26) + (1 * 26) + 1, "AAB"),
        ((17 * 26 * 26) + (5 * 26) + 3, "QED"),
    ],
)
def test_index_to_col(index, expected_col):
    assert index_to_col(index) == expected_col


@pytest.mark.parametrize(
    "col1, col2, expected_result",
    [
        ("A", "B", False),
        ("B", "A", True),
        ("AA", "AB", False),
        ("AB", "AA", True),
        ("A", "AA", False),
        ("AA", "A", True),
        ("Z", "AA", False),
        ("AA", "Z", True),
        ("AAA", "AAB", False),
        ("AAB", "AAA", True),
        ("QED", "QEE", False),
        ("QEE", "QED", True),
    ],
)
def test_is_col_greater(col1, col2, expected_result):
    assert is_col_greater(col1, col2) == expected_result


def test_compute_sheet_data_dimensions(sheet_data_input_fixture):
    (min_row, max_row), (min_col_index, max_col_index) = compute_sheet_data_dimensions(
        sheet_data_input_fixture
    )

    expected_min_row = 1
    expected_max_row = 44
    expected_min_col_index = 0  # Column "A"
    expected_max_col_index = 8  # Column "I"

    assert min_row == expected_min_row
    assert max_row == expected_max_row
    assert min_col_index == expected_min_col_index
    assert max_col_index == expected_max_col_index


def test_create_sheet_properties():
    sheet_properties = create_sheet_properties(
        sheet_id=1,
        title="Sheet1",
        row_count=10000,
        column_count=260,
    )

    assert sheet_properties.sheetId == 1
    assert sheet_properties.title == "Sheet1"
    assert sheet_properties.gridProperties.rowCount == 10000
    assert sheet_properties.gridProperties.columnCount == 260


@pytest.mark.parametrize(
    "row_numbers, expected_groups",
    [
        ([], []),
        ([5, 6, 7], [[5, 6, 7]]),
        (
            [1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 18, 19, 20],
            [[1, 2, 3], [5, 6, 7, 8, 9, 10, 11], [18, 19, 20]],
        ),
    ],
)
def test_group_contiguous_rows(row_numbers, expected_groups):
    grouped_rows = group_contiguous_rows(row_numbers)
    assert grouped_rows == expected_groups


@pytest.mark.parametrize(
    "input_value, expected_key, expected_value, expected_type, expected_pattern",
    [
        (1234, "numberValue", 1234, NumberFormatType.NUMBER, "#,##0"),
        (1.234, "numberValue", 1.234, NumberFormatType.NUMBER, "#,##0.00"),
        ("$100", "numberValue", 100, NumberFormatType.CURRENCY, "$#,##0"),
        ("$100.50", "numberValue", 100.50, NumberFormatType.CURRENCY, "$#,##0.00"),
        ("75%", "numberValue", 75.00, NumberFormatType.PERCENT, "0.00%"),
        ("75.34%", "numberValue", 75.34, NumberFormatType.PERCENT, "0.00%"),
        ("$1abc", "stringValue", "$1abc", None, None),
        ("abc7%", "stringValue", "abc7%", None, None),
        ("=SUM(A1:B1)", "formulaValue", "=SUM(A1:B1)", None, None),
        (True, "boolValue", True, None, None),
    ],
)
def test_create_cell_data(
    input_value, expected_key, expected_value, expected_type, expected_pattern
):
    cell_data = create_cell_data(input_value)
    expected_cell_value = CellExtendedValue(**{expected_key: expected_value})
    assert cell_data.userEnteredValue == expected_cell_value
    if expected_type is None:
        assert cell_data.userEnteredFormat is None
    else:
        assert cell_data.userEnteredFormat is not None
        assert cell_data.userEnteredFormat.numberFormat.type == expected_type
        assert cell_data.userEnteredFormat.numberFormat.pattern == expected_pattern


def test_create_row_data():
    row_data = {
        "A": 1,  # Column index 0
        "B": 2.5,  # Column index 1
        "AA": "test",  # Column index 26
        "BA": True,  # Column index 52
        "BB": "=SUM(A1:B1)",  # Column index 53
    }
    min_col_index = 0  # Column "A"
    max_col_index = 53  # Column "BB"

    expected_row_data = RowData(
        values=[
            CellData(userEnteredValue=CellExtendedValue(stringValue=""))
            for _ in range(max_col_index + 1)
        ]
    )
    expected_row_data.values[0].userEnteredValue = CellExtendedValue(numberValue=1)
    expected_row_data.values[1].userEnteredValue = CellExtendedValue(numberValue=2.5)
    expected_row_data.values[26].userEnteredValue = CellExtendedValue(stringValue="test")
    expected_row_data.values[52].userEnteredValue = CellExtendedValue(boolValue=True)
    expected_row_data.values[53].userEnteredValue = CellExtendedValue(formulaValue="=SUM(A1:B1)")

    row_data = create_row_data(row_data, min_col_index, max_col_index)

    assert len(row_data.values) == len(expected_row_data.values)
    for cell, expected in zip(row_data.values, expected_row_data.values, strict=False):
        assert cell.userEnteredValue == expected.userEnteredValue


def test_create_sheet_data():
    from arcade_google.models import CellData, CellExtendedValue, SheetDataInput
    from arcade_google.utils import create_cell_data

    test_data = {
        2: {"B": "row2B", "C": 200},
        3: {"B": "row3B"},
        5: {"A": "=SUM(A1:A1)", "C": "row5C"},
    }
    sheet_data_input = SheetDataInput(data=test_data)
    min_col_index = 0  # Column "A"
    max_col_index = 2  # Column "C"

    grid_data_list = create_sheet_data(sheet_data_input, min_col_index, max_col_index)

    assert len(grid_data_list) == 2, "Should have two groups of contiguous rows"

    group1 = grid_data_list[0]
    assert group1.startRow == 1
    assert group1.startColumn == min_col_index
    assert len(group1.rowData) == 2

    row2_cells = group1.rowData[0].values
    expected_row2 = [
        CellData(userEnteredValue=CellExtendedValue(stringValue="")),
        create_cell_data("row2B"),
        create_cell_data(200),
    ]
    for cell, expected in zip(row2_cells, expected_row2, strict=False):
        assert cell.userEnteredValue == expected.userEnteredValue

    row3_cells = group1.rowData[1].values
    expected_row3 = [
        CellData(userEnteredValue=CellExtendedValue(stringValue="")),
        create_cell_data("row3B"),
        CellData(userEnteredValue=CellExtendedValue(stringValue="")),
    ]
    for cell, expected in zip(row3_cells, expected_row3, strict=False):
        assert cell.userEnteredValue == expected.userEnteredValue

    group2 = grid_data_list[1]
    assert group2.startRow == 4
    assert group2.startColumn == min_col_index
    assert len(group2.rowData) == 1

    row5_cells = group2.rowData[0].values
    expected_row5 = [
        create_cell_data("=SUM(A1:A1)"),
        CellData(userEnteredValue=CellExtendedValue(stringValue="")),
        create_cell_data("row5C"),
    ]
    for cell, expected in zip(row5_cells, expected_row5, strict=False):
        assert cell.userEnteredValue == expected.userEnteredValue


@pytest.mark.parametrize(
    "cell, expected",
    [
        ({}, ""),
        ({"userEnteredValue": {}}, ""),
        ({"userEnteredValue": {"stringValue": "hello"}}, "hello"),
        ({"userEnteredValue": {"numberValue": 123}}, 123),
        ({"userEnteredValue": {"boolValue": True}}, True),
        ({"userEnteredValue": {"formulaValue": "=SUM(A1:A2)"}}, "=SUM(A1:A2)"),
    ],
)
def test_extract_user_entered_cell_value(cell, expected):
    result = extract_user_entered_cell_value(cell)
    assert result == expected


def test_process_row_empty():
    row = {}
    assert process_row(row, 0) == {}


def test_process_row_non_empty():
    row = {
        "values": [
            {"userEnteredValue": {"stringValue": "cell1"}, "formattedValue": "cell1"},
            {"userEnteredValue": {}},  # should be ignored
            {"userEnteredValue": {"formulaValue": "=C1+D4"}, "formattedValue": 42},
            {"userEnteredValue": {"stringValue": ""}, "formattedValue": ""},  # should be ignored
            {"userEnteredValue": {"boolValue": False}, "formattedValue": False},
        ]
    }
    expected = {
        "A": {"userEnteredValue": "cell1", "formattedValue": "cell1"},
        "C": {"userEnteredValue": "=C1+D4", "formattedValue": 42},
        "E": {"userEnteredValue": False, "formattedValue": False},
    }

    assert process_row(row, 0) == expected


def test_process_row_with_start_index():
    row = {
        "values": [
            {"userEnteredValue": {"stringValue": "x"}, "formattedValue": "x"},
            {"userEnteredValue": {"formulaValue": "=C1+D4"}, "formattedValue": "$10.00"},
        ]
    }
    expected = {
        "C": {"userEnteredValue": "x", "formattedValue": "x"},
        "D": {"userEnteredValue": "=C1+D4", "formattedValue": "$10.00"},
    }

    assert process_row(row, 2) == expected


def test_convert_api_grid_data_to_dict_single_grid():
    data = [
        {
            "startRow": 0,
            "startColumn": 0,
            "rowData": [
                {
                    "values": [
                        {"userEnteredValue": {"stringValue": "A1"}, "formattedValue": "A1"},
                        {"userEnteredValue": {"numberValue": 1}, "formattedValue": 1},
                    ]
                },
                {
                    "values": [
                        {"userEnteredValue": {"stringValue": "A2"}, "formattedValue": "A2"},
                        {"userEnteredValue": {"numberValue": 2}, "formattedValue": 2},
                    ]
                },
                {
                    "values": [
                        {"userEnteredValue": {}},
                        {
                            "userEnteredValue": {"stringValue": "ignored"},
                            "formattedValue": "ignored",
                        },
                        {"userEnteredValue": {"numberValue": 3}, "formattedValue": 3},
                    ]
                },
            ],
        }
    ]
    result = convert_api_grid_data_to_dict(data)
    expected = {
        1: {
            "A": {"userEnteredValue": "A1", "formattedValue": "A1"},
            "B": {"userEnteredValue": 1, "formattedValue": 1},
        },
        2: {
            "A": {"userEnteredValue": "A2", "formattedValue": "A2"},
            "B": {"userEnteredValue": 2, "formattedValue": 2},
        },
        3: {
            "B": {"userEnteredValue": "ignored", "formattedValue": "ignored"},
            "C": {"userEnteredValue": 3, "formattedValue": 3},
        },
    }

    assert result == expected


def test_convert_api_grid_data_to_dict_multiple_grids():
    data = [
        {
            "startRow": 5,
            "startColumn": 1,
            "rowData": [
                {
                    "values": [
                        {"userEnteredValue": {"numberValue": 100}, "formattedValue": 100},
                        {"userEnteredValue": {"stringValue": "=SUM(A1:A2)"}, "formattedValue": 23},
                    ]
                }
            ],
        },
        {
            "startRow": 0,
            "startColumn": 0,
            "rowData": [
                {
                    "values": [
                        {"userEnteredValue": {"stringValue": "First"}, "formattedValue": "First"},
                        {"userEnteredValue": {"numberValue": 10}, "formattedValue": 10},
                    ]
                }
            ],
        },
    ]
    result = convert_api_grid_data_to_dict(data)
    expected = {
        1: {
            "A": {"userEnteredValue": "First", "formattedValue": "First"},
            "B": {"userEnteredValue": 10, "formattedValue": 10},
        },
        6: {
            "B": {"userEnteredValue": 100, "formattedValue": 100},
            "C": {"userEnteredValue": "=SUM(A1:A2)", "formattedValue": 23},
        },
    }

    assert result == expected


def test_convert_api_grid_data_to_dict_empty_rows():
    data = [
        {
            "startRow": 10,
            "startColumn": 0,
            "rowData": [
                {"values": [{"userEnteredValue": {}, "formattedValue": ""}]},
                {"values": []},
            ],
        }
    ]
    result = convert_api_grid_data_to_dict(data)
    expected = {}

    assert result == expected


FAKE_SHEET_RESPONSE = {
    "sheets": [
        {"properties": {"title": "Sheet1", "gridProperties": {"rowCount": 10, "columnCount": 6}}}
    ]
}


@patch("arcade_google.utils.build_sheets_service")
def test_validate_write_to_cell_params_valid(mock_build):
    mock_service = MagicMock()
    mock_service.spreadsheets().get().execute.return_value = FAKE_SHEET_RESPONSE
    mock_build.return_value = mock_service

    service = mock_build("dummy_token")

    validate_write_to_cell_params(
        service=service,
        spreadsheet_id="dummy_id",
        sheet_name="Sheet1",
        column="B",
        row=5,
    )


@patch("arcade_google.utils.build_sheets_service")
def test_validate_write_to_cell_params_invalid_sheet_name(mock_build):
    mock_service = MagicMock()
    mock_service.spreadsheets().get().execute.return_value = FAKE_SHEET_RESPONSE
    mock_build.return_value = mock_service

    service = mock_build("dummy_token")

    with pytest.raises(RetryableToolError) as excinfo:
        validate_write_to_cell_params(
            service=service,
            spreadsheet_id="dummy_id",
            sheet_name="NonExistentSheet",
            column="A",
            row=5,
        )
    assert "Sheet name NonExistentSheet not found" in str(excinfo.value)


@patch("arcade_google.utils.build_sheets_service")
def test_validate_write_to_cell_params_row_out_of_bounds(mock_build):
    mock_service = MagicMock()
    mock_service.spreadsheets().get().execute.return_value = FAKE_SHEET_RESPONSE
    mock_build.return_value = mock_service

    service = mock_build("dummy_token")

    out_of_bounds_row = 15
    with pytest.raises(ToolExecutionError) as excinfo:
        validate_write_to_cell_params(
            service=service,
            spreadsheet_id="dummy_id",
            sheet_name="Sheet1",
            column="A",
            row=out_of_bounds_row,
        )
    assert f"Row {out_of_bounds_row} is out of bounds" in str(excinfo.value)


@patch("arcade_google.utils.build_sheets_service")
def test_validate_write_to_cell_params_column_out_of_bounds(mock_build):
    mock_service = MagicMock()
    mock_service.spreadsheets().get().execute.return_value = FAKE_SHEET_RESPONSE
    mock_build.return_value = mock_service

    service = mock_build("dummy_token")

    out_of_bounds_column = "Z"
    with pytest.raises(ToolExecutionError) as excinfo:
        validate_write_to_cell_params(
            service=service,
            spreadsheet_id="dummy_id",
            sheet_name="Sheet1",
            column=out_of_bounds_column,
            row=5,
        )
    assert f"Column {out_of_bounds_column} is out of bounds" in str(excinfo.value)



from typing import Any

from arcade_tdk import ToolContext
from arcade_tdk.errors import ToolExecutionError


def get_tweet_url(tweet_id: str) -> str:
    """Get the URL of a tweet given its ID."""
    return f"https://x.com/x/status/{tweet_id}"


def get_headers_with_token(context: ToolContext) -> dict[str, str]:
    """Get the headers for a request to the X API."""
    if context.authorization is None or context.authorization.token is None:
        raise ToolExecutionError(
            "Missing Token. Authorization is required to post a tweet.",
            developer_message="Token is not set in the ToolContext.",
        )
    token = (
        context.authorization.token if context.authorization and context.authorization.token else ""
    )
    return {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
    }


def parse_search_recent_tweets_response(response_data: dict[str, Any]) -> dict[str, Any]:
    """
    Parses response from the X API search recent tweets endpoint.
    Returns the modified response data with added 'tweet_url', 'author_username', and 'author_name'.
    """
    if not sanity_check_tweets_data(response_data):
        return {"data": [], "next_token": ""}

    # Add 'tweet_url' to each tweet
    for tweet in response_data["data"]:
        tweet["tweet_url"] = get_tweet_url(tweet["id"])

    # Add 'author_username' and 'author_name' to each tweet
    for tweet_data, user_data in zip(
        response_data["data"], response_data["includes"]["users"], strict=False
    ):
        tweet_data["author_username"] = user_data["username"]
        tweet_data["author_name"] = user_data["name"]

    return response_data


def sanity_check_tweets_data(tweets_data: dict[str, Any]) -> bool:
    """
    Sanity check the tweets data.
    Returns True if the tweets data is valid and contains tweets, False otherwise.
    """
    if not tweets_data.get("data"):
        return False
    # prefer clarity over appeasing linter here
    if not tweets_data.get("includes", {}).get("users"):  # noqa: SIM103
        return False
    return True


def expand_long_tweet(tweet_data: dict[str, Any]) -> None:
    """Expand a long tweet.

    For tweets exceeding 280 characters,
    replace the truncated tweet text with the full tweet text.
    """
    if tweet_data.get("note_tweet"):
        tweet_data["text"] = tweet_data["note_tweet"]["text"]
        del tweet_data["note_tweet"]


def expand_urls_in_tweets(
    tweets_data: list[dict[str, Any]], delete_entities: bool = True
) -> list[dict[str, Any]]:
    """
    Returns a new list of tweets with expanded URLs.
    """
    new_tweets = []
    for tweet_data in tweets_data:
        new_tweet = tweet_data.copy()
        if "entities" in new_tweet and "urls" in new_tweet["entities"]:
            for url_entity in new_tweet["entities"]["urls"]:
                short_url = url_entity["url"]
                expanded_url = url_entity["expanded_url"]
                new_tweet["text"] = new_tweet["text"].replace(short_url, expanded_url)

        if delete_entities:
            new_tweet.pop("entities", None)
        new_tweets.append(new_tweet)
    return new_tweets


def expand_urls_in_user_description(user_data: dict, delete_entities: bool = True) -> dict:
    """
    Returns a new user data dict with expanded URLs in the description.
    """
    new_user_data = user_data.copy()
    description_urls = new_user_data.get("entities", {}).get("description", {}).get("urls", [])
    description = new_user_data.get("description", "")
    for url_info in description_urls:
        t_co_link = url_info["url"]
        expanded_url = url_info["expanded_url"]
        description = description.replace(t_co_link, expanded_url)
    new_user_data["description"] = description

    if delete_entities:
        new_user_data.pop("entities", None)
    return new_user_data


def expand_urls_in_user_url(user_data: dict, delete_entities: bool = True) -> dict:
    """
    Returns a new user data dict with expanded URLs in the URL field.
    """
    new_user_data = user_data.copy()
    url_urls = new_user_data.get("entities", {}).get("url", {}).get("urls", [])
    url = new_user_data.get("url", "")
    for url_info in url_urls:
        t_co_link = url_info["url"]
        expanded_url = url_info["expanded_url"]
        url = url.replace(t_co_link, expanded_url)
    new_user_data["url"] = url

    if delete_entities:
        new_user_data.pop("entities", None)
    return new_user_data


def remove_none_values(params: dict) -> dict:
    """
    Remove key/value pairs with None values from a dictionary.

    Args:
        params: The dictionary to clean

    Returns:
        A new dictionary with None values removed
    """
    return {k: v for k, v in params.items() if v is not None}


def expand_attached_media(params: dict) -> dict:
    """
    Include attached media metadata in the request parameters.
    """
    params["expansions"] += ",attachments.media_keys"
    params["tweet.fields"] += ",attachments"
    params["media.fields"] = ",".join([
        # media_key, url and type are returned by default, added here for clarity
        "media_key",
        "url",
        "type",
        "duration_ms",
        "height",
        "width",
        "preview_image_url",
        "alt_text",
        "public_metrics",
    ])
    return params



from arcade_google_calendar.tools.calendar import (
    create_event,
    delete_event,
    find_time_slots_when_everyone_is_free,
    list_calendars,
    list_events,
    update_event,
)

__all__ = [
    "create_event",
    "delete_event",
    "find_time_slots_when_everyone_is_free",
    "list_calendars",
    "list_events",
    "update_event",
]



class GmailToolError(Exception):
    """Base exception for Google tool errors."""

    def __init__(self, message: str, developer_message: str | None = None):
        self.message = message
        self.developer_message = developer_message
        super().__init__(self.message)

    def __str__(self) -> str:
        base_message = self.message
        if self.developer_message:
            return f"{base_message} (Developer: {self.developer_message})"
        return base_message


class GmailServiceError(GmailToolError):
    """Raised when there's an error building or using the Google service."""

    pass



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from arcade_tdk import ToolContext

from arcade_google.tools import create_contact


@pytest.fixture
def mock_context():
    context = AsyncMock(spec=ToolContext)
    context.authorization = MagicMock()
    context.authorization.token = "mock_token"  # noqa: S105
    return context


@pytest.mark.asyncio
async def test_create_contact_success(mock_context):
    # Test create_contact with all parameters (given, family names and email)
    created_contact_data = {"resourceName": "people/123", "etag": "abc"}

    create_contact_call = MagicMock()
    create_contact_call.execute.return_value = created_contact_data

    people_mock = MagicMock()
    people_mock.createContact.return_value = create_contact_call

    service_mock = MagicMock()
    service_mock.people.return_value = people_mock

    with patch(
        "arcade_google.tools.contacts.build_people_service", return_value=service_mock
    ) as mock_build:
        result = await create_contact(
            mock_context,
            given_name="Alice",
            family_name="Smith",
            email="alice@example.com",
        )
        assert "contact" in result
        assert result["contact"] == created_contact_data

        # Verify that the createContact API was called with the correct body contents.
        expected_body = {
            "names": [{"givenName": "Alice", "familyName": "Smith"}],
            "emailAddresses": [{"value": "alice@example.com", "type": "work"}],
        }
        people_mock.createContact.assert_called_once_with(
            body=expected_body, personFields="names,emailAddresses"
        )
        mock_build.assert_called_once()


@pytest.mark.asyncio
async def test_create_contact_success_without_optional(mock_context):
    # Test create_contact without optional parameters family_name and email.
    created_contact_data = {"resourceName": "people/456", "etag": "def"}

    create_contact_call = MagicMock()
    create_contact_call.execute.return_value = created_contact_data

    people_mock = MagicMock()
    people_mock.createContact.return_value = create_contact_call

    service_mock = MagicMock()
    service_mock.people.return_value = people_mock

    with patch("arcade_google.tools.contacts.build_people_service", return_value=service_mock):
        result = await create_contact(mock_context, given_name="Bob", family_name=None, email=None)
        assert "contact" in result
        assert result["contact"] == created_contact_data

        # Expected body should only include the givenName when family_name and email are omitted.
        expected_body = {"names": [{"givenName": "Bob"}]}
        people_mock.createContact.assert_called_once_with(
            body=expected_body, personFields="names,emailAddresses"
        )


@pytest.mark.asyncio
async def test_create_contact_error(mock_context):
    # Simulate an error thrown by createContact
    error_call = MagicMock()
    error_call.execute.side_effect = Exception("Create error")

    people_mock = MagicMock()
    people_mock.createContact.return_value = error_call

    service_mock = MagicMock()
    service_mock.people.return_value = people_mock

    with (
        patch("arcade_google.tools.contacts.build_people_service", return_value=service_mock),
        pytest.raises(Exception, match="Error in execution of CreateContact"),
    ):
        await create_contact(mock_context, given_name="Alice", family_name="Doe", email=None)



import pytest
from arcade_cli.utils import compute_base_url, compute_login_url

DEFAULT_CLOUD_HOST = "cloud.arcade.dev"
DEFAULT_ENGINE_HOST = "api.arcade.dev"
LOCALHOST = "localhost"
DEFAULT_PORT = None
DEFAULT_FORCE_TLS = False
DEFAULT_FORCE_NO_TLS = False


@pytest.mark.parametrize(
    "inputs, expected_output",
    [
        pytest.param(
            {
                "host_input": DEFAULT_ENGINE_HOST,
                "port_input": DEFAULT_PORT,
                "force_tls": DEFAULT_FORCE_TLS,
                "force_no_tls": DEFAULT_FORCE_NO_TLS,
            },
            "https://api.arcade.dev",
            id="default",
        ),
        pytest.param(
            {
                "host_input": LOCALHOST,
                "port_input": DEFAULT_PORT,
                "force_tls": DEFAULT_FORCE_TLS,
                "force_no_tls": DEFAULT_FORCE_NO_TLS,
            },
            "http://localhost:9099",
            id="localhost",
        ),
        pytest.param(
            {
                "host_input": DEFAULT_ENGINE_HOST,
                "port_input": 9099,
                "force_tls": DEFAULT_FORCE_TLS,
                "force_no_tls": DEFAULT_FORCE_NO_TLS,
            },
            "https://api.arcade.dev:9099",
            id="custom port",
        ),
        pytest.param(
            {
                "host_input": LOCALHOST,
                "port_input": 9099,
                "force_tls": DEFAULT_FORCE_TLS,
                "force_no_tls": DEFAULT_FORCE_NO_TLS,
            },
            "http://localhost:9099",
            id="localhost with custom port",
        ),
        pytest.param(
            {
                "host_input": DEFAULT_ENGINE_HOST,
                "port_input": DEFAULT_PORT,
                "force_tls": True,
                "force_no_tls": DEFAULT_FORCE_NO_TLS,
            },
            "https://api.arcade.dev",
            id="force TLS",
        ),
        pytest.param(
            {
                "host_input": LOCALHOST,
                "port_input": DEFAULT_PORT,
                "force_tls": True,
                "force_no_tls": DEFAULT_FORCE_NO_TLS,
            },
            "https://localhost:9099",
            id="localhost with force TLS",
        ),
        pytest.param(
            {
                "host_input": DEFAULT_ENGINE_HOST,
                "port_input": 9099,
                "force_tls": True,
                "force_no_tls": DEFAULT_FORCE_NO_TLS,
            },
            "https://api.arcade.dev:9099",
            id="custom port with force TLS",
        ),
        pytest.param(
            {
                "host_input": LOCALHOST,
                "port_input": 9099,
                "force_tls": True,
                "force_no_tls": DEFAULT_FORCE_NO_TLS,
            },
            "https://localhost:9099",
            id="localhost with custom port and force TLS",
        ),
        pytest.param(
            {
                "host_input": DEFAULT_ENGINE_HOST,
                "port_input": DEFAULT_PORT,
                "force_tls": DEFAULT_FORCE_TLS,
                "force_no_tls": True,
            },
            "http://api.arcade.dev",
            id="force no TLS",
        ),
        pytest.param(
            {
                "host_input": LOCALHOST,
                "port_input": DEFAULT_PORT,
                "force_tls": DEFAULT_FORCE_TLS,
                "force_no_tls": True,
            },
            "http://localhost:9099",
            id="localhost with force no TLS",
        ),
        pytest.param(
            {
                "host_input": DEFAULT_ENGINE_HOST,
                "port_input": 9099,
                "force_tls": DEFAULT_FORCE_TLS,
                "force_no_tls": True,
            },
            "http://api.arcade.dev:9099",
            id="custom port with force no TLS",
        ),
        pytest.param(
            {
                "host_input": LOCALHOST,
                "port_input": 9099,
                "force_tls": DEFAULT_FORCE_TLS,
                "force_no_tls": True,
            },
            "http://localhost:9099",
            id="localhost with custom port and force no TLS",
        ),
        pytest.param(
            {
                "host_input": DEFAULT_ENGINE_HOST,
                "port_input": DEFAULT_PORT,
                "force_tls": True,
                "force_no_tls": True,
            },
            "http://api.arcade.dev",
            id="force TLS and no TLS",
        ),
        pytest.param(
            {
                "host_input": LOCALHOST,
                "port_input": DEFAULT_PORT,
                "force_tls": True,
                "force_no_tls": True,
            },
            "http://localhost:9099",
            id="localhost with force TLS and no TLS",
        ),
        pytest.param(
            {
                "host_input": DEFAULT_ENGINE_HOST,
                "port_input": 9099,
                "force_tls": True,
                "force_no_tls": True,
            },
            "http://api.arcade.dev:9099",
            id="custom port with force TLS and no TLS",
        ),
        pytest.param(
            {
                "host_input": LOCALHOST,
                "port_input": 9099,
                "force_tls": True,
                "force_no_tls": True,
            },
            "http://localhost:9099",
            id="localhost with custom port, force TLS and no TLS",
        ),
        pytest.param(
            {
                "host_input": "arandomhost.com",
                "port_input": DEFAULT_PORT,
                "force_tls": DEFAULT_FORCE_TLS,
                "force_no_tls": DEFAULT_FORCE_NO_TLS,
            },
            "https://arandomhost.com",
            id="random host",
        ),
    ],
)
def test_compute_base_url(inputs: dict, expected_output: str):
    base_url = compute_base_url(
        inputs["force_tls"],
        inputs["force_no_tls"],
        inputs["host_input"],
        inputs["port_input"],
    )

    assert base_url == expected_output


@pytest.mark.parametrize(
    "inputs, expected_output",
    [
        pytest.param(
            {"host_input": DEFAULT_CLOUD_HOST, "port_input": DEFAULT_PORT, "state": "123"},
            "https://cloud.arcade.dev/api/v1/auth/cli_login?callback_uri=http%3A%2F%2Flocalhost%3A9905%2Fcallback&state=123",
            id="default",
        ),
        pytest.param(
            {"host_input": "localhost", "port_input": 9099, "state": "123"},
            "http://localhost:9099/api/v1/auth/cli_login?callback_uri=http%3A%2F%2Flocalhost%3A9905%2Fcallback&state=123",
            id="localhost with custom port",
        ),
        pytest.param(
            {"host_input": "localhost", "port_input": DEFAULT_PORT, "state": "123"},
            "http://localhost:8000/api/v1/auth/cli_login?callback_uri=http%3A%2F%2Flocalhost%3A9905%2Fcallback&state=123",
            id="localhost",
        ),
        pytest.param(
            {"host_input": DEFAULT_CLOUD_HOST, "port_input": 8000, "state": "123"},
            "https://cloud.arcade.dev/api/v1/auth/cli_login?callback_uri=http%3A%2F%2Flocalhost%3A9905%2Fcallback&state=123",
            id="cloud host with an ignored custom port",
        ),
        pytest.param(
            {
                "host_input": DEFAULT_CLOUD_HOST,
                "port_input": DEFAULT_PORT,
                "state": "123",
                "callback_host": "other-host.com/123",
            },
            "https://cloud.arcade.dev/api/v1/auth/cli_login?callback_uri=http%3A%2F%2Fother-host.com%2F123%2Fcallback&state=123",
            id="cloud host with a custom callback host",
        ),
    ],
)
def test_compute_login_url(inputs: dict, expected_output: str):
    callback_host = inputs.get("callback_host")
    login_url = compute_login_url(
        inputs["host_input"], inputs["state"], inputs["port_input"], callback_host
    )

    assert login_url == expected_output



from dataclasses import dataclass
from typing import Any

from arcade_evals.critic import BinaryCritic


@dataclass
class HasSubstringCritic(BinaryCritic):
    """A critic for checking whether the argument value contains a substring."""

    def evaluate(self, expected: Any, actual: Any) -> dict[str, float | bool]:
        """
        Evaluates whether the actual value contains the expected value.

        Args:
            expected: The expected value.
            actual: The actual value to compare, cast to the type of expected.

        Returns:
            dict: A dictionary containing the match status and score.
        """
        if not isinstance(actual, str) or not isinstance(expected, str):
            return {"match": False, "score": 0.0}

        try:
            actual_casted = self.cast_actual(expected, actual)
        except TypeError:
            actual_casted = actual

        match = expected in actual_casted
        return {"match": match, "score": self.weight if match else 0.0}


@dataclass
class CaseInsensitiveBinaryCritic(BinaryCritic):
    """A critic for checking whether actual and expected values are the same, case insensitive."""

    def evaluate(self, expected: Any, actual: Any) -> dict[str, float | bool]:
        """
        Evaluates whether the actual value is the same as the expected value, case insensitive.

        Args:
            expected: The expected value.
            actual: The actual value to compare, cast to the type of expected.

        Returns:
            dict: A dictionary containing the match status and score.
        """
        if not isinstance(actual, str) or not isinstance(expected, str):
            return {"match": False, "score": 0.0}

        try:
            actual_casted = self.cast_actual(expected, actual)
        except TypeError:
            actual_casted = actual

        match = expected.casefold() in actual_casted.casefold()
        return {"match": match, "score": self.weight if match else 0.0}


@dataclass
class CaseInsensitiveListOfStringsBinaryCritic(BinaryCritic):
    """Checks that all strings match in Actual and Expected list of strings, case insensitive"""

    def evaluate(self, expected: Any, actual: Any) -> dict[str, float | bool]:
        """
        Checks that all strings match in Actual and Expected list of strings, case insensitive

        Args:
            expected: The expected value.
            actual: The actual value to compare, cast to the type of expected.

        Returns:
            dict: A dictionary containing the match status and score.
        """
        if not isinstance(actual, list) or not isinstance(expected, list):
            return {"match": False, "score": 0.0}

        all_actual_str = all(isinstance(item, str) for item in actual)
        all_expected_str = all(isinstance(item, str) for item in expected)

        if not all_actual_str or not all_expected_str:
            return {"match": False, "score": 0.0}

        actual_folded = [item.casefold() for item in actual]
        expected_folded = [item.casefold() for item in expected]

        match = len(actual) == len(expected) and all(
            item in actual_folded for item in expected_folded
        )
        return {"match": match, "score": self.weight if match else 0.0}



from collections.abc import Callable

import pytest
from arcade_tdk import ToolContext

from arcade_jira.exceptions import JiraToolExecutionError
from arcade_jira.tools.priorities import list_projects_associated_with_a_priority_scheme
from arcade_jira.utils import add_pagination_to_response, clean_project_dict, paginate_all_items


@pytest.mark.asyncio
async def test_paginate_all_items_with_zero_matches(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_search_response_dict: Callable,
):
    response = mock_httpx_response(200, build_project_search_response_dict([], is_last=True))
    mock_httpx_client.get.return_value = response

    response = await paginate_all_items(
        context=mock_context,
        tool=list_projects_associated_with_a_priority_scheme,
        response_items_key="projects",
        scheme_id="123",
    )
    assert response == []


@pytest.mark.asyncio
async def test_paginate_all_items_with_one_page(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    build_project_search_response_dict: Callable,
    fake_cloud_name: str,
):
    projects = [build_project_dict(), build_project_dict()]
    response = mock_httpx_response(200, build_project_search_response_dict(projects, is_last=True))
    mock_httpx_client.get.return_value = response

    response = await paginate_all_items(
        context=mock_context,
        tool=list_projects_associated_with_a_priority_scheme,
        response_items_key="projects",
        scheme_id="123",
    )
    assert response == [clean_project_dict(project, fake_cloud_name) for project in projects]


@pytest.mark.asyncio
async def test_paginate_all_items_with_multiple_pages(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    build_project_search_response_dict: Callable,
    fake_cloud_name: str,
):
    page1 = [build_project_dict(), build_project_dict()]
    page2 = [build_project_dict(), build_project_dict()]
    page3 = [build_project_dict()]

    response1 = mock_httpx_response(200, build_project_search_response_dict(page1, is_last=False))
    response2 = mock_httpx_response(200, build_project_search_response_dict(page2, is_last=False))
    response3 = mock_httpx_response(200, build_project_search_response_dict(page3, is_last=True))

    mock_httpx_client.get.side_effect = [response1, response2, response3]

    response = await paginate_all_items(
        context=mock_context,
        tool=list_projects_associated_with_a_priority_scheme,
        response_items_key="projects",
        scheme_id="123",
        limit=2,
    )
    assert response == [
        clean_project_dict(project, fake_cloud_name) for project in page1 + page2 + page3
    ]


@pytest.mark.asyncio
async def test_paginate_all_items_when_tool_returns_error(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_search_response_dict: Callable,
):
    project_id = "456"
    get_project_by_id_response = mock_httpx_response(404, {})
    search_projects_response = mock_httpx_response(200, build_project_search_response_dict([]))
    mock_httpx_client.get.side_effect = [get_project_by_id_response, search_projects_response]

    with pytest.raises(JiraToolExecutionError) as exc:
        await paginate_all_items(
            context=mock_context,
            tool=list_projects_associated_with_a_priority_scheme,
            response_items_key="projects",
            scheme_id="123",
            project=project_id,
            limit=2,
        )

    assert exc.value.message == f"Project not found with name/key/ID '{project_id}'"


def test_add_pagination_to_response_with_zero_items():
    response = {"items": []}
    items = []
    limit = 10
    offset = 0
    add_pagination_to_response(response, items, limit, offset)
    assert response["pagination"] == {"is_last_page": True, "limit": limit, "total_results": 0}


def test_add_pagination_to_response_with_last_page_false():
    items = [{"id": "123"}, {"id": "456"}]
    response = {"items": items, "isLast": False}
    limit = 2
    offset = 0
    add_pagination_to_response(response, items, limit, offset)
    assert response["pagination"] == {
        "limit": limit,
        "total_results": 2,
        "next_offset": 2,
    }


def test_add_pagination_to_response_with_last_page_true():
    items = [{"id": "123"}, {"id": "456"}]
    response = {"items": items, "isLast": True}
    limit = 2
    offset = 0
    add_pagination_to_response(response, items, limit, offset)
    assert response["pagination"] == {
        "limit": limit,
        "total_results": 2,
        "is_last_page": True,
    }


def test_add_pagination_to_response_without_last_page_and_limit_equal_items():
    items = [{"id": "123"}, {"id": "456"}]
    response = {"items": items}
    limit = 2
    offset = 0
    add_pagination_to_response(response, items, limit, offset)
    assert response["pagination"] == {
        "limit": limit,
        "total_results": 2,
        "next_offset": 2,
    }


def test_add_pagination_to_response_without_last_page_and_less_items_than_limit():
    items = [{"id": "123"}]
    response = {"items": items}
    limit = 2
    offset = 0
    add_pagination_to_response(response, items, limit, offset)
    assert response["pagination"] == {
        "limit": limit,
        "total_results": 1,
        "is_last_page": True,
    }



from arcade_evals import (
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

from arcade_google_drive.tools import (
    get_file_tree_structure,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
catalog.add_tool(get_file_tree_structure, "GoogleDrive")

get_file_tree_structure_history = [
    {"role": "system", "content": "Today is 2025-07-03, Thursday."},
    {"role": "user", "content": "get my file tree structure"},
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_EnqRPmIx3zrquDA7PuZ5NtK6",
                "type": "function",
                "function": {"name": "GoogleDrive_GetFileTreeStructure", "arguments": "{}"},
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"drives":[]}',
        "tool_call_id": "call_EnqRPmIx3zrquDA7PuZ5NtK6",
        "name": "GoogleDrive_GetFileTreeStructure",
    },
    {
        "role": "assistant",
        "content": "I could not find any files in your Google Drive. To select and grant access permissions to additional files and folders, use [this Google Drive File Picker link](https://cloud.bosslevel.dev/api/v1/google/drive_picker?config=eyXRoIjogM3NsdnFrMmtqODlldNuLmF0=).",
    },
]


@tool_eval()
def tools_understand_filepicker_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Google Drive tools."""
    suite = EvalSuite(
        name="Google Drive Tools Evaluation",
        system_message="You are an AI assistant that can manage Google Drive using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Ensure LLM understands that after using file picker, it should call the tool again",
        user_message="ok i followed your suggestion and went to that url. how has it changed?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_file_tree_structure,
                args={},
            )
        ],
        additional_messages=get_file_tree_structure_history,
    )
    return suite



from enum import Enum


class BodyFormat(str, Enum):
    STORAGE = "storage"  # Storage representation for editing, with relative urls in the markup

    def to_api_value(self) -> str:
        mapping = {
            BodyFormat.STORAGE: "storage",
        }
        return mapping.get(self, BodyFormat.STORAGE.value)


class PageUpdateMode(str, Enum):
    """The mode of update for a page"""

    PREPEND = "prepend"  # Add content to the beginning of the page.
    APPEND = "append"  # Add content to the end of the page.
    REPLACE = "replace"  # Replace the entire page with the new content.


class PageSortOrder(str, Enum):
    """The order of the pages to sort by"""

    ID_ASCENDING = "id-ascending"
    ID_DESCENDING = "id-descending"
    TITLE_ASCENDING = "title-ascending"
    TITLE_DESCENDING = "title-descending"
    CREATED_DATE_ASCENDING = "created-date-oldest-to-newest"
    CREATED_DATE_DESCENDING = "created-date-newest-to-oldest"
    MODIFIED_DATE_ASCENDING = "modified-date-oldest-to-newest"
    MODIFIED_DATE_DESCENDING = "modified-date-newest-to-oldest"

    def to_api_value(self) -> str:
        mapping = {
            PageSortOrder.ID_ASCENDING: "id",
            PageSortOrder.ID_DESCENDING: "-id",
            PageSortOrder.TITLE_ASCENDING: "title",
            PageSortOrder.TITLE_DESCENDING: "-title",
            PageSortOrder.CREATED_DATE_ASCENDING: "created-date",
            PageSortOrder.CREATED_DATE_DESCENDING: "-created-date",
            PageSortOrder.MODIFIED_DATE_ASCENDING: "modified-date",
            PageSortOrder.MODIFIED_DATE_DESCENDING: "-modified-date",
        }
        return mapping.get(self, PageSortOrder.CREATED_DATE_DESCENDING.value)


class AttachmentSortOrder(str, Enum):
    """The order of the attachments to sort by"""

    CREATED_DATE_ASCENDING = "created-date-oldest-to-newest"
    CREATED_DATE_DESCENDING = "created-date-newest-to-oldest"
    MODIFIED_DATE_ASCENDING = "modified-date-oldest-to-newest"
    MODIFIED_DATE_DESCENDING = "modified-date-newest-to-oldest"

    def to_api_value(self) -> str:
        mapping = {
            AttachmentSortOrder.CREATED_DATE_ASCENDING: "created-date",
            AttachmentSortOrder.CREATED_DATE_DESCENDING: "-created-date",
            AttachmentSortOrder.MODIFIED_DATE_ASCENDING: "modified-date",
            AttachmentSortOrder.MODIFIED_DATE_DESCENDING: "-modified-date",
        }
        return mapping.get(self, AttachmentSortOrder.CREATED_DATE_DESCENDING.value)



# Base URL for GitHub API
GITHUB_API_BASE_URL = "https://api.github.com"

# Endpoint patterns
ENDPOINTS = {
    "repo": "/repos/{owner}/{repo}",
    "org_repos": "/orgs/{org}/repos",
    "repo_activity": "/repos/{owner}/{repo}/activity",
    "repo_pulls_comments": "/repos/{owner}/{repo}/pulls/comments",
    "repo_issues": "/repos/{owner}/{repo}/issues",
    "repo_issue_comments": "/repos/{owner}/{repo}/issues/{issue_number}/comments",
    "repo_pulls": "/repos/{owner}/{repo}/pulls",
    "repo_pull": "/repos/{owner}/{repo}/pulls/{pull_number}",
    "repo_pull_commits": "/repos/{owner}/{repo}/pulls/{pull_number}/commits",
    "repo_pull_comments": "/repos/{owner}/{repo}/pulls/{pull_number}/comments",
    "repo_pull_comment_replies": "/repos/{owner}/{repo}/pulls/{pull_number}/comments/{comment_id}/replies",  # noqa: E501
    "user_starred": "/user/starred/{owner}/{repo}",
    "repo_stargazers": "/repos/{owner}/{repo}/stargazers",
}



from enum import Enum


class SortDirection(str, Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"


class ObjectType(str, Enum):
    PAGE = "page"
    DATABASE = "database"


class BlockType(str, Enum):
    BOOKMARK = "bookmark"
    BREADCRUMB = "breadcrumb"
    BULLETED_LIST_ITEM = "bulleted_list_item"
    CALLOUT = "callout"
    CHILD_DATABASE = "child_database"
    CHILD_PAGE = "child_page"
    COLUMN = "column"
    COLUMN_LIST = "column_list"
    DIVIDER = "divider"
    EMBED = "embed"
    EQUATION = "equation"
    FILE = "file"
    HEADING_1 = "heading_1"
    HEADING_2 = "heading_2"
    HEADING_3 = "heading_3"
    IMAGE = "image"
    LINK_PREVIEW = "link_preview"
    LINK_TO_PAGE = "link_to_page"
    NUMBERED_LIST_ITEM = "numbered_list_item"
    PARAGRAPH = "paragraph"
    PDF = "pdf"
    QUOTE = "quote"
    SYNCED_BLOCK = "synced_block"
    TABLE = "table"
    TABLE_OF_CONTENTS = "table_of_contents"
    TABLE_ROW = "table_row"
    TEMPLATE = "template"
    TO_DO = "to_do"
    TOGGLE = "toggle"
    UNSUPPORTED = "unsupported"
    VIDEO = "video"



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    SimilarityCritic,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_notion_toolkit
from arcade_notion_toolkit.tools import (
    create_page,
    get_page_content_by_id,
    get_page_content_by_title,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_notion_toolkit)

PAGE_CONTENT = """## Why Build Tools with Arcade?

Arcade solves key challenges for agent developers:

1. **Auth Native to Agents**: Authentication designed for agentic workflows — the right token is always available for each user without complex integration work.

2. **Multi-Tenant Tool Calling**: Enable your agent to take actions AS the specific user of the agent

3. **Better Agent Capabilities**: Build tools that securely connect to the services your users want your agent to integrate with (Gmail, Slack, Google Drive, Zoom, etc.) without complex integration code.

4. **Clean Codebase**: Eliminate environment variables full of API keys and complex OAuth implementations from your application code.

5. **Flexible Integration**: Choose your integration approach:

    - LLM API for the simplest experience with hundreds of pre-built tools
    - Tools API for direct execution control
    - Auth API for authentication-only integration
    - Framework connectors for LangChain, CrewAI and others

6. **Zero Schema Maintenance**: Tool definitions generate automatically from code annotations and translate to any LLM format.

7. **Built-in Evaluation**: Evaluate your tools across user scenarios, llms, and context with Arcade's tool calling evaluation framework. Ensure your tools are working as expected and are useful for your agents.

8. **Complete Tooling Ecosystem**: Built-in evaluation framework, scalable execution infrastructure, and flexible deployment options (including VPC, Docker, and Kubernetes).

Arcade lets you focus on creating useful tool functionality rather than solving complex authentication, deployment, and integration challenges.
"""  # noqa: E501

# A conversation where a user asks the AI to get the content of a page named 'Arcade Notes'
GET_PAGE_CONVERSATION = [
    {"role": "user", "content": "get 'Arcade Notes'"},
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_92Vhl75I8KEKQfjihS7l53DL",
                "type": "function",
                "function": {
                    "name": "Notion_GetPageContentByTitle",
                    "arguments": '{"title":"Arcade Notes"}',
                },
            }
        ],
    },
    {
        "role": "tool",
        "content": "# Arcade Notes\nFirst, make sure you have these pre-requisites installed on your system:  \n- **Python 3.10**\xa0or higherVerify your Python version by running\xa0`python --version`\xa0or\xa0`python3 --version`\xa0in your terminal.  \n- **pip**: The Python package installer should be available. It's typically included with Python.  \n- **Arcade Account**: Sign up for an\xa0[Arcade account](https://api.arcade.dev/signup?utm_source=docs&utm_medium=page&utm_campaign=custom-tools)\xa0if you haven't already.  \nLet's set up Arcade and give it a try!  \n### **Obtain an API key**  \nInstall the Arcade CLI and SDK and log in. Your Arcade API key will be printed to the console as well as saved to\xa0`~/.arcade/credentials.yaml`.  \npip install arcade-aiarcade login\n### **Try\xa0**`arcade chat`  \nWith Arcade CLI installed, you can test outour API with the\xa0`arcade chat`\xa0command:  \narcade chat\nThis launches a chat with the Arcade Cloud Engine (hosted at\xa0`api.arcade.dev`). All pre-built Arcade tools are available to use.  \nFor example, try asking:  \nstar the ArcadeAI/arcade-ai repo on Github\nArcade will ask you to authorize with GitHub, and then the AI assistant will star the\xa0[ArcadeAI/arcade-ai](https://github.com/ArcadeAI/arcade-ai)\xa0repo on your behalf.  \nYou'll see output similar to this:  \nAssistant (gpt-4o):I starred the ArcadeAI/arcade-ai repo on Github for you!\nYou can use Ctrl-C to exit the chat at any time.  \n### **Try more tools**  \nTry prompting the AI with different requests:  \n- Search the web to find when The Winds of Winter will be released  \n- Summarize my emails from today  \n- What's playing on my Spotify right now?  \nTo see all tools available in the Cloud, run:  \narcade show\nGreat job! 🎉 You've set up Arcade and used your first tool.  \n### **Next steps**  \n- [Learn about tools](https://docs.arcade.dev/home/use-tools/tools-overview)\xa0andhow they work  \n- Learn how to\xa0[call tools with models](https://docs.arcade.dev/home/use-tools/call-tools-with-models)  \n- Or, jump ahead to building your own\xa0[custom tools](https://docs.arcade.dev/home/build-tools/create-a-toolkit)  \n",  # noqa: E501
        "tool_call_id": "call_92Vhl75I8KEKQfjihS7l53DL",
        "name": "Notion_GetPageContentByTitle",
    },
    {
        "role": "assistant",
        "content": "Here are the notes on setting up and using Arcade:\n\n### Pre-requisites\nMake sure you have the following installed on your system:\n-**Python 3.10 or higher**: Verify using `python --version` or `python3 --version` in your terminal.\n- **pip**: The Python package installer, typically included with Python.\n- **Arcade Account**: Sign up for an [Arcade account](https://api.arcade.dev/signup?utm_source=docs&utm_medium=page&utm_campaign=custom-tools) if you haven't already.\n\n### Setting up Arcade\n\n#### Obtain an API key\n1. Install the Arcade CLI and SDK.\n2. Log in to Arcade: Run the command `arcade login`.\n3. Your Arcade API key will be printed to the console and saved in `~/.arcade/credentials.yaml`.\n\n#### Try `arcade chat`\n- With Arcade CLI installed, you can test the API using `arcade chat`.\n- This command launches a chat with the Arcade Cloud Engine (hosted at `api.arcade.dev`), giving you access to all pre-built Arcade tools.\n  \nExample prompt:  \n- Ask to star the `ArcadeAI/arcade-ai` repo on GitHub. The AI assistant will star the repo for you after authorization.\n\nEvent example output:\n```\nAssistant (gpt-4o): I starred the ArcadeAI/arcade-ai repo on Github for you!\n```\n\nUse `Ctrl-C` to exit the chat anytime.\n\n### Explore More Tools\n- Try different requests like searching the web, summarizing emails, or checking your current Spotify play.\n- Run `arcade show` to see all available tools in the Cloud.\n\n### Next Steps\n- [Learn about tools](https://docs.arcade.dev/home/use-tools/tools-overview) and how they function.\n- Understand how to [call tools with models](https://docs.arcade.dev/home/use-tools/call-tools-with-models).\n- Begin building your own [custom tools](https://docs.arcade.dev/home/build-tools/create-a-toolkit).\n\nCongratulations on setting up Arcade and trying your first tool! 🎉",  # noqa: E501
    },
]


@tool_eval()
def create_page_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools creating a Notion page"""
    suite = EvalSuite(
        name="Notion Create Page Evaluation",
        system_message=(
            "You are an AI assistant that has access to the user's Notion workspace. "
            "You can take actions on the user's Notion workspace on behalf of the user."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    # Easy case
    suite.add_case(
        name="Create page easy difficulty",
        user_message=(
            "Create a page with the title '07/11/2027' and content '* drank a slurpie' "
            "under the parent page 'Daily Standup'."
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_page,
                args={
                    "parent_title": "Daily Standup",
                    "title": "07/11/2027",
                    "content": "* drank a slurpie",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="parent_title", weight=0.34),
            SimilarityCritic(critic_field="title", weight=0.33, similarity_threshold=0.95),
            SimilarityCritic(critic_field="content", weight=0.33, similarity_threshold=0.95),
        ],
    )

    # Medium case
    suite.add_case(
        name="Create page medium difficulty",
        user_message=(
            f"Create a page with the title 'Why Use Arcade?' and content {PAGE_CONTENT}"
            "under the parent page 'Arcade Notes'."
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_page,
                args={
                    "parent_title": "Arcade Notes",
                    "title": "Why Use Arcade?",
                    "content": PAGE_CONTENT,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="parent_title", weight=0.34),
            SimilarityCritic(critic_field="title", weight=0.33, similarity_threshold=0.95),
            SimilarityCritic(critic_field="content", weight=0.33, similarity_threshold=0.95),
        ],
    )

    # Hard case
    suite.add_case(
        name="Create page hard difficulty",
        user_message=(f"Add {PAGE_CONTENT} as a subpage. Name it 'Why Use Arcade?'"),
        expected_tool_calls=[
            ExpectedToolCall(
                func=create_page,
                args={
                    "parent_title": "Arcade Notes",
                    "title": "Why Use Arcade?",
                    "content": PAGE_CONTENT,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="parent_title", weight=0.34),
            SimilarityCritic(critic_field="title", weight=0.33, similarity_threshold=0.95),
            SimilarityCritic(critic_field="content", weight=0.33, similarity_threshold=0.95),
        ],
        additional_messages=GET_PAGE_CONVERSATION,
    )
    return suite


@tool_eval()
def get_page_content_eval_suite() -> EvalSuite:
    """Create an evaluation suite for tools getting the content of a Notion page"""
    suite = EvalSuite(
        name="Notion Get Page Content By ID Evaluation",
        system_message=(
            "You are an AI assistant that has access to the user's Notion workspace. "
            "You can take actions on the user's Notion workspace on behalf of the user."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    # Easy case
    suite.add_case(
        name="Get page content by id easy difficulty",
        user_message="Get the content of the page with id 1b37a62b04d48079a902ce69ed7e7240",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_page_content_by_id,
                args={
                    "page_id": "1b37a62b04d48079a902ce69ed7e7240",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="page_id", weight=1),
        ],
    )

    # Medium case
    suite.add_case(
        name="Get page content medium difficulty",
        user_message=(
            "Summarize the main points in 1b37a62b04d48079a902ce69ed7e7240. "
            "Also, does 'Tool Calling with Arcade' actually talk about tools?"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_page_content_by_id,
                args={
                    "page_id": "1b37a62b04d48079a902ce69ed7e7240",
                },
            ),
            ExpectedToolCall(
                func=get_page_content_by_title,
                args={
                    "title": "Tool Calling with Arcade",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="page_id", weight=0.5),
            BinaryCritic(critic_field="title", weight=0.5),
        ],
    )

    # Hard case
    suite.add_case(
        name="Get page content hard difficulty",
        user_message=(
            "Compare it's main points against 'Tool Calling with Arcade' and "
            "'Tool Execution with Arcade'"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_page_content_by_title,
                args={
                    "title": "Tool Calling with Arcade",
                },
            ),
            ExpectedToolCall(
                func=get_page_content_by_title,
                args={
                    "title": "Tool Execution with Arcade",
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="title", weight=1),
        ],
        additional_messages=GET_PAGE_CONVERSATION,
    )

    return suite



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Microsoft
from arcade_tdk.errors import ToolExecutionError

from arcade_microsoft.client import get_client
from arcade_microsoft.outlook_mail._utils import (
    fetch_emails,
    prepare_list_emails_request_config,
    remove_none_values,
)
from arcade_microsoft.outlook_mail.enums import (
    EmailFilterProperty,
    FilterOperator,
    WellKnownFolderNames,
)
from arcade_microsoft.outlook_mail.message import Message


@tool(requires_auth=Microsoft(scopes=["Mail.Read"]))
async def list_emails(
    context: ToolContext,
    limit: Annotated[int, "The number of messages to return. Max is 100. Defaults to 5."] = 5,
    pagination_token: Annotated[
        str | None, "The pagination token to continue a previous request"
    ] = None,
) -> Annotated[dict, "A dictionary containing a list of emails"]:
    """List emails in the user's mailbox across all folders.

    Since this tool lists email across all folders, it may return sent items, drafts,
    and other items that are not in the inbox.
    """
    client = get_client(context.get_auth_token_or_empty())
    request_config = prepare_list_emails_request_config(limit)
    message_builder = client.me.messages

    response = await fetch_emails(message_builder, pagination_token, request_config)
    messages = [Message.from_sdk(msg).to_dict() for msg in response.value or []]
    pagination_token = response.odata_next_link

    result = {
        "messages": messages,
        "num_messages": len(messages),
        "pagination_token": pagination_token,
    }
    result = remove_none_values(result)
    return result


@tool(requires_auth=Microsoft(scopes=["Mail.Read"]))
async def list_emails_in_folder(
    context: ToolContext,
    well_known_folder_name: Annotated[
        WellKnownFolderNames | None,
        "The name of the folder to list emails from. Defaults to None.",
    ] = None,
    folder_id: Annotated[
        str | None,
        "The ID of the folder to list emails from if the folder is not a well-known folder. "
        "Defaults to None.",
    ] = None,
    limit: Annotated[int, "The number of messages to return. Max is 100. Defaults to 5."] = 5,
    pagination_token: Annotated[
        str | None, "The pagination token to continue a previous request"
    ] = None,
) -> Annotated[
    dict, "A dictionary containing a list of emails and a pagination token, if applicable"
]:
    """List the user's emails in the specified folder.

    Exactly one of `well_known_folder_name` or `folder_id` MUST be provided.
    """
    if not (bool(well_known_folder_name) ^ bool(folder_id)):
        raise ToolExecutionError(
            message="Exactly one of `well_known_folder_name` or `folder_id` must be provided."
        )
    folder_name = well_known_folder_name.value if well_known_folder_name else folder_id
    client = get_client(context.get_auth_token_or_empty())
    request_config = prepare_list_emails_request_config(limit)
    message_builder = client.me.mail_folders.by_mail_folder_id(folder_name).messages  # type: ignore[arg-type]

    response = await fetch_emails(message_builder, pagination_token, request_config)
    messages = [Message.from_sdk(msg).to_dict() for msg in response.value or []]
    pagination_token = response.odata_next_link

    result = {
        "messages": messages,
        "num_messages": len(messages),
        "pagination_token": pagination_token,
    }
    result = remove_none_values(result)
    return result


@tool(requires_auth=Microsoft(scopes=["Mail.Read"]))
async def list_emails_by_property(
    context: ToolContext,
    property: Annotated[EmailFilterProperty, "The property to filter the emails by."],  # noqa: A002
    operator: Annotated[FilterOperator, "The operator to use for the filter."],
    value: Annotated[str, "The value to filter the emails by"],
    limit: Annotated[int, "The number of messages to return. Max is 100. Defaults to 5."] = 5,
    pagination_token: Annotated[
        str | None, "The pagination token to continue a previous request"
    ] = None,
) -> Annotated[dict, "A dictionary containing a list of emails"]:
    """List emails in the user's mailbox across all folders filtering by a property."""
    client = get_client(context.get_auth_token_or_empty())
    request_config = prepare_list_emails_request_config(limit, property, operator, value)
    message_builder = client.me.messages

    response = await fetch_emails(message_builder, pagination_token, request_config)
    messages = [Message.from_sdk(msg).to_dict() for msg in response.value or []]
    pagination_token = response.odata_next_link

    result = {
        "messages": messages,
        "num_messages": len(messages),
        "pagination_token": pagination_token,
    }
    result = remove_none_values(result)
    return result



import asyncio
from typing import Annotated, Any, cast

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Atlassian

import arcade_jira.cache as cache
from arcade_jira.client import JiraClient
from arcade_jira.constants import JIRA_API_REQUEST_TIMEOUT, PrioritySchemeOrderBy
from arcade_jira.exceptions import JiraToolExecutionError, MultipleItemsFoundError, NotFoundError
from arcade_jira.utils import (
    add_pagination_to_response,
    clean_priority_dict,
    clean_priority_scheme_dict,
    clean_project_dict,
    find_priorities_by_project,
    find_unique_project,
    remove_none_values,
)


@tool(requires_auth=Atlassian(scopes=["read:jira-work"]))
async def get_priority_by_id(
    context: ToolContext,
    priority_id: Annotated[str, "The ID of the priority to retrieve."],
) -> Annotated[dict[str, Any], "The priority"]:
    """Get the details of a priority by its ID."""
    client = JiraClient(context.get_auth_token_or_empty())
    try:
        response = await client.get(f"/priority/{priority_id}")
    except NotFoundError:
        return {"error": f"Priority not found with id '{priority_id}'"}
    return {"priority": clean_priority_dict(response)}


@tool(requires_auth=Atlassian(scopes=["manage:jira-configuration"]))
async def list_priority_schemes(
    context: ToolContext,
    scheme_name: Annotated[
        str | None, "Filter by scheme name. Defaults to None (returns all scheme names)."
    ] = None,
    limit: Annotated[
        int,
        "The maximum number of priority schemes to return. Min of 1, max of 50. Defaults to 50.",
    ] = 50,
    offset: Annotated[
        int, "The number of priority schemes to skip. Defaults to 0 (start from the first scheme)."
    ] = 0,
    order_by: Annotated[
        PrioritySchemeOrderBy,
        "The order in which to return the priority schemes. Defaults to name ascending.",
    ] = PrioritySchemeOrderBy.NAME_ASCENDING,
) -> Annotated[dict[str, Any], "The priority schemes available"]:
    """Browse the priority schemes available in Jira."""
    limit = max(min(limit, 50), 1)
    client = JiraClient(context.get_auth_token_or_empty())
    api_response = await client.get(
        "/priorityscheme",
        params=remove_none_values({
            "startAt": offset,
            "maxResults": limit,
            "schemeName": scheme_name,
            "orderBy": order_by.to_api_value(),
        }),
    )
    cloud_name = cache.get_cloud_name(context.get_auth_token_or_empty())
    schemes = [clean_priority_scheme_dict(scheme, cloud_name) for scheme in api_response["values"]]
    response = {
        "priority_schemes": schemes,
        "isLast": api_response.get("isLast"),
    }
    return add_pagination_to_response(response, schemes, limit, offset)


@tool(requires_auth=Atlassian(scopes=["manage:jira-configuration"]))
async def list_priorities_associated_with_a_priority_scheme(
    context: ToolContext,
    scheme_id: Annotated[str, "The ID of the priority scheme to retrieve priorities for."],
    limit: Annotated[
        int,
        "The maximum number of priority schemes to return. Min of 1, max of 50. Defaults to 50.",
    ] = 50,
    offset: Annotated[
        int, "The number of priority schemes to skip. Defaults to 0 (start from the first scheme)."
    ] = 0,
) -> Annotated[dict[str, Any], "The priorities associated with the priority scheme"]:
    """Browse the priorities associated with a priority scheme."""
    client = JiraClient(context.get_auth_token_or_empty())
    api_response = await client.get(
        f"/priorityscheme/{scheme_id}/priorities",
        params={
            "startAt": offset,
            "maxResults": limit,
        },
    )
    priorities = [clean_priority_dict(priority) for priority in api_response["values"]]
    response = {
        "priorities": priorities,
        "isLast": api_response.get("isLast"),
    }
    return add_pagination_to_response(response, priorities, limit, offset)


@tool(requires_auth=Atlassian(scopes=["manage:jira-configuration"]))
async def list_projects_associated_with_a_priority_scheme(
    context: ToolContext,
    scheme_id: Annotated[str, "The ID of the priority scheme to retrieve projects for."],
    project: Annotated[
        str | None, "Filter by project ID, key or name. Defaults to None (returns all projects)."
    ] = None,
    limit: Annotated[
        int,
        "The maximum number of projects to return. Min of 1, max of 50. Defaults to 50.",
    ] = 50,
    offset: Annotated[
        int, "The number of projects to skip. Defaults to 0 (start from the first project)."
    ] = 0,
) -> Annotated[dict[str, Any], "The projects associated with the priority scheme"]:
    """Browse the projects associated with a priority scheme."""
    if project:
        try:
            project_data = await find_unique_project(context, project)
        except (NotFoundError, MultipleItemsFoundError) as exc:
            return {"error": exc.message}
        else:
            project = project_data["id"]

    client = JiraClient(context.get_auth_token_or_empty())
    api_response = await client.get(
        f"/priorityscheme/{scheme_id}/projects",
        params=remove_none_values({
            "startAt": offset,
            "maxResults": limit,
            "projectId": project,
        }),
    )
    cloud_name = cache.get_cloud_name(context.get_auth_token_or_empty())
    projects = [clean_project_dict(project, cloud_name) for project in api_response["values"]]
    response = {
        "projects": projects,
        "isLast": api_response.get("isLast"),
    }
    return add_pagination_to_response(response, projects, limit, offset)


@tool(requires_auth=Atlassian(scopes=["manage:jira-configuration"]))
async def list_priorities_available_to_a_project(
    context: ToolContext,
    project: Annotated[str, "The ID, key or name of the project to retrieve priorities for."],
) -> Annotated[
    dict[str, Any],
    "The priorities available to be used in issues in the specified Jira project",
]:
    """Browse the priorities available to be used in issues in the specified Jira project.

    This tool may need to loop through several API calls to get all priorities associated with
    a specific project. In Jira environments with too many Projects or Priority Schemes,
    the search may take too long, and the tool call will timeout.
    """
    try:
        project_data = await find_unique_project(context, project)
    except (NotFoundError, MultipleItemsFoundError) as exc:
        return {"error": exc.message}

    try:
        return await asyncio.wait_for(
            find_priorities_by_project(context, project_data),
            timeout=JIRA_API_REQUEST_TIMEOUT,
        )
    except asyncio.TimeoutError:
        return {"error": f"The operation timed out after {JIRA_API_REQUEST_TIMEOUT} seconds."}
    except JiraToolExecutionError as error:
        return {"error": error.message}


@tool(requires_auth=Atlassian(scopes=["manage:jira-configuration"]))
async def list_priorities_available_to_an_issue(
    context: ToolContext,
    issue: Annotated[str, "The ID or key of the issue to retrieve priorities for."],
) -> Annotated[dict[str, Any], "The priorities available to be used in the specified Jira issue"]:
    """Browse the priorities available to be used in the specified Jira issue."""
    from arcade_jira.tools.issues import get_issue_by_id

    issue_response = await get_issue_by_id(context, issue)
    if issue_response.get("error"):
        return cast(dict[str, Any], issue_response)

    issue_data = issue_response["issue"]
    project = issue_data["project"]["id"]

    response = await list_priorities_available_to_a_project(context, project)

    return {
        "issue": {
            "id": issue_data["id"],
            "key": issue_data["key"],
            "title": issue_data["title"],
        },
        "project": response["project"],
        "priorities_available": response["priorities_available"],
    }



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_dropbox
from arcade_dropbox.critics import DropboxPathCritic
from arcade_dropbox.tools.browse import list_items_in_folder

rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
catalog.add_module(arcade_dropbox)


@tool_eval()
def list_items_in_folder_eval_suite() -> EvalSuite:
    """Create an evaluation suite for the list_items_in_folder tool."""
    suite = EvalSuite(
        name="list_items_in_folder",
        system_message="You are an AI assistant that can interact with files and folders in Dropbox using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="List items in the Dropbox root folder",
        user_message="List the items in the Dropbox root folder",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_items_in_folder,
                args={
                    "folder_path": "",
                    "limit": 100,
                    "cursor": None,
                },
            ),
        ],
        critics=[
            DropboxPathCritic(critic_field="folder_path", weight=0.6),
            BinaryCritic(critic_field="limit", weight=0.2),
            BinaryCritic(critic_field="cursor", weight=0.2),
        ],
    )

    suite.add_case(
        name="List items in a sub-folder",
        user_message="List the items in the folder AcmeInc/Reports",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_items_in_folder,
                args={
                    "folder_path": "/AcmeInc/Reports",
                    "limit": 100,
                    "cursor": None,
                },
            ),
        ],
        critics=[
            DropboxPathCritic(critic_field="folder_path", weight=0.6),
            BinaryCritic(critic_field="limit", weight=0.2),
            BinaryCritic(critic_field="cursor", weight=0.2),
        ],
    )

    suite.add_case(
        name="List items in a sub-folder with custom limit",
        user_message="List the first 50 items in the folder AcmeInc/Reports",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_items_in_folder,
                args={
                    "folder_path": "/AcmeInc/Reports",
                    "limit": 50,
                    "cursor": None,
                },
            ),
        ],
        critics=[
            DropboxPathCritic(critic_field="folder_path", weight=0.4),
            BinaryCritic(critic_field="limit", weight=0.4),
            BinaryCritic(critic_field="cursor", weight=0.2),
        ],
    )

    return suite



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Atlassian

from arcade_confluence.client import ConfluenceClientV1


@tool(
    requires_auth=Atlassian(
        scopes=["search:confluence"],
    )
)
async def search_content(
    context: ToolContext,
    must_contain_all: Annotated[
        list[str] | None,
        "Words/phrases that content MUST contain (AND logic). Each item can be:\n"
        "- Single word: 'banana' - content must contain this word\n"
        "- Multi-word phrase: 'How to' - content must contain all these words (in any order)\n"
        "- All items in this list must be present for content to match\n"
        "- Example: ['banana', 'apple'] finds content containing BOTH 'banana' AND 'apple'",
    ] = None,
    can_contain_any: Annotated[
        list[str] | None,
        "Words/phrases where content can contain ANY of these (OR logic). Each item can be:\n"
        "- Single word: 'project' - content containing this word will match\n"
        "- Multi-word phrase: 'pen & paper' - content containing all these words will match\n"
        "- Content matching ANY item in this list will be included\n"
        "- Example: ['project', 'documentation'] finds content with 'project' OR 'documentation'",
    ] = None,
    enable_fuzzy: Annotated[
        bool,
        "Enable fuzzy matching to find similar terms (e.g. 'roam' will find 'foam'). "
        "Defaults to True",
    ] = True,
    limit: Annotated[int, "Maximum number of results to return (1-100). Defaults to 25"] = 25,
) -> Annotated[dict, "Search results containing content items matching the criteria"]:
    """Search for content in Confluence.

    The search is performed across all content in the authenticated user's Confluence workspace.
    All search terms in Confluence are case insensitive.

    You can use the parameters in different ways:
    - must_contain_all: For AND logic - content must contain ALL of these
    - can_contain_any: For OR logic - content can contain ANY of these
    - Combine them: must_contain_all=['banana'] AND can_contain_any=['database', 'guide']
    """
    client = ConfluenceClientV1(context.get_auth_token_or_empty())
    cql = client.construct_cql(must_contain_all, can_contain_any, enable_fuzzy)
    response = await client.get("search", params={"cql": cql, "limit": max(1, min(limit, 100))})

    return client.transform_search_content_response(response)



from arcade_reddit.tools.read import (
    check_subreddit_access,
    get_content_of_multiple_posts,
    get_content_of_post,
    get_my_posts,
    get_my_username,
    get_posts_in_subreddit,
    get_subreddit_rules,
    get_top_level_comments,
)
from arcade_reddit.tools.submit import (
    comment_on_post,
    reply_to_comment,
    submit_text_post,
)

__all__ = [
    "check_subreddit_access",
    "comment_on_post",
    "get_content_of_multiple_posts",
    "get_content_of_post",
    "get_my_posts",
    "get_my_username",
    "get_posts_in_subreddit",
    "get_subreddit_rules",
    "get_top_level_comments",
    "reply_to_comment",
    "submit_text_post",
]



from arcade_asana.tools.projects import get_project_by_id, list_projects
from arcade_asana.tools.tags import create_tag, list_tags
from arcade_asana.tools.tasks import (
    attach_file_to_task,
    create_task,
    get_subtasks_from_a_task,
    get_task_by_id,
    get_tasks_without_id,
    update_task,
)
from arcade_asana.tools.teams import get_team_by_id, list_teams_the_current_user_is_a_member_of
from arcade_asana.tools.users import get_user_by_id, list_users
from arcade_asana.tools.workspaces import list_workspaces

__all__ = [
    "attach_file_to_task",
    "create_tag",
    "create_task",
    "get_project_by_id",
    "get_subtasks_from_a_task",
    "get_task_by_id",
    "get_team_by_id",
    "get_user_by_id",
    "list_projects",
    "list_tags",
    "list_teams_the_current_user_is_a_member_of",
    "list_users",
    "list_workspaces",
    "get_tasks_without_id",
    "update_task",
]



import asyncio
import logging
import os
import uuid
from enum import Enum
from typing import Any, Callable, Union

from arcade_core.catalog import MaterializedTool, ToolCatalog
from arcade_core.executor import ToolExecutor
from arcade_core.schema import ToolAuthorizationContext, ToolContext
from arcadepy import ArcadeError, AsyncArcade
from arcadepy.types.auth_authorize_params import AuthRequirement, AuthRequirementOauth2
from arcadepy.types.shared import AuthorizationResponse

from arcade_serve.mcp.convert import convert_to_mcp_content, create_mcp_tool
from arcade_serve.mcp.logging import create_mcp_logging_middleware
from arcade_serve.mcp.message_processor import MCPMessageProcessor, create_message_processor
from arcade_serve.mcp.types import (
    CallToolRequest,
    CallToolResponse,
    CallToolResult,
    CancelRequest,
    Implementation,
    InitializeRequest,
    InitializeResponse,
    InitializeResult,
    JSONRPCError,
    JSONRPCResponse,
    ListPromptsRequest,
    ListPromptsResponse,
    ListResourcesRequest,
    ListResourcesResponse,
    ListToolsRequest,
    ListToolsResponse,
    ListToolsResult,
    PingRequest,
    PingResponse,
    ProgressNotification,
    ServerCapabilities,
    ShutdownRequest,
    ShutdownResponse,
    Tool,
)

logger = logging.getLogger("arcade.mcp")

MCP_PROTOCOL_VERSION = "2024-11-05"


class MessageMethod(str, Enum):
    """Enumeration of supported MCP message methods"""

    PING = "ping"
    INITIALIZE = "initialize"
    LIST_TOOLS = "tools/list"
    CALL_TOOL = "tools/call"
    PROGRESS = "progress"
    CANCEL = "$/cancelRequest"
    SHUTDOWN = "shutdown"
    LIST_RESOURCES = "resources/list"
    LIST_PROMPTS = "prompts/list"


class MCPServer:
    """
    Unified async MCP server that manages connections, middleware, and tool invocation.
    Handles protocol-level messages (ping, initialize, list_tools, call_tool, etc.).
    """

    def __init__(
        self,
        tool_catalog: Any,
        enable_logging: bool = True,
        **client_kwargs: dict[str, Any],
    ) -> None:
        """
        Initialize the MCP server.

        Args:
            tool_catalog: Catalog of available tools
            **client_kwargs: Additional arguments to pass to the AsyncArcade client
        """
        self.tool_catalog: ToolCatalog = tool_catalog
        self.message_processor: MCPMessageProcessor = create_message_processor()

        # Pop middleware_config from client_kwargs regardless of logging state,
        # as it's internal config not meant for AsyncArcade.
        middleware_config = client_kwargs.pop("middleware_config", {})

        if enable_logging:
            # Create and add the logging middleware if logging is enabled.
            # Note: enable_logging must be True for this middleware (and its stdio_mode behavior)
            # to be activated.
            self.message_processor.add_middleware(
                create_mcp_logging_middleware(**middleware_config)
            )

        self._shutdown: bool = False
        # Initialize AsyncArcade with the *remaining* client_kwargs
        self.arcade = AsyncArcade(**client_kwargs)  # type: ignore[arg-type]

        # Initialize handler dispatch table
        self._method_handlers: dict[str, Callable] = {
            MessageMethod.PING: self._handle_ping,
            MessageMethod.INITIALIZE: self._handle_initialize,
            MessageMethod.LIST_TOOLS: self._handle_list_tools,
            MessageMethod.CALL_TOOL: self._handle_call_tool,
            MessageMethod.PROGRESS: self._handle_progress,
            MessageMethod.CANCEL: self._handle_cancel,
            MessageMethod.SHUTDOWN: self._handle_shutdown,
            MessageMethod.LIST_RESOURCES: self._handle_list_resources,
            MessageMethod.LIST_PROMPTS: self._handle_list_prompts,
        }

    async def run_connection(
        self,
        read_stream: Any,
        write_stream: Any,
        init_options: Any,
    ) -> None:
        """
        Handle a single MCP connection (SSE or stdio).

        Args:
            read_stream: Async iterable yielding incoming messages.
            write_stream: Object with an async send(message) method.
            init_options: Initialization options for the connection.
        """
        # Generate a user ID if possible
        user_id = self._get_user_id(init_options)

        try:
            logger.info(f"Starting MCP connection for user {user_id}")

            async for message in read_stream:
                # Process the message
                response = await self.handle_message(message, user_id=user_id)

                # Skip sending responses for None (e.g., notifications)
                if response is None:
                    continue

                await self._send_response(write_stream, response)

        except asyncio.CancelledError:
            logger.info("Connection cancelled")
        except Exception:
            logger.exception("Error in connection")

    def _get_user_id(self, init_options: Any) -> str:
        """
        Get the user ID for a connection.

        Args:
            init_options: Initialization options for the connection

        Returns:
            A user ID string
        """
        try:
            from arcade_core.config import config

            # Prefer config.user.email if available
            if config.user and config.user.email:
                return config.user.email
        except ValueError:
            logger.debug("No logged in user for MCP Server")

        fallback = str(uuid.uuid4())
        if os.environ.get("ARCADE_USER_ID", None):
            return os.environ.get("ARCADE_USER_ID", fallback)
        elif isinstance(init_options, dict):
            user_id = init_options.get("user_id")
            if user_id:
                return str(user_id)
        # Fallback to random UUID
        return str(fallback)

    async def _send_response(self, write_stream: Any, response: Any) -> None:
        """
        Send a response to the client.

        Args:
            write_stream: Stream to write the response to
            response: Response object to send
        """
        # Ensure the response is properly serialized to JSON
        if hasattr(response, "model_dump_json"):
            # It's a Pydantic model, serialize it
            json_response = response.model_dump_json()
            # Ensure it ends with a newline for JSON-RPC-over-stdio
            if not json_response.endswith("\n"):
                json_response += "\n"
            logger.debug(f"Sending response: {json_response[:200]}...")
            await write_stream.send(json_response)
        elif isinstance(response, dict):
            # It's a dict, convert to JSON
            import json

            json_response = json.dumps(response)
            # Ensure it ends with a newline for JSON-RPC-over-stdio
            if not json_response.endswith("\n"):
                json_response += "\n"
            logger.debug(f"Sending response: {json_response[:200]}...")
            await write_stream.send(json_response)
        else:
            # It's already a string or something else
            response_str = str(response)
            # Ensure it ends with a newline for JSON-RPC-over-stdio
            if not response_str.endswith("\n"):
                response_str += "\n"
            logger.debug(f"Sending raw response type: {type(response)}")
            await write_stream.send(response_str)

    async def handle_message(self, message: Any, user_id: str | None = None) -> Any:
        """
        Handle an incoming MCP message. Processes it through middleware and dispatches
        to the appropriate handler based on the message method.

        Args:
            message: The raw incoming message
            user_id: Optional user ID for authentication

        Returns:
            A properly formatted response message
        """
        # Pre-process message through middleware
        processed = await self.message_processor.process_request(message)

        # Handle special case for JSON string initialize requests
        if isinstance(processed, str):
            try:
                import json

                parsed = json.loads(processed)
                if (
                    isinstance(parsed, dict)
                    and parsed.get("method") == MessageMethod.INITIALIZE
                    and "id" in parsed
                ):
                    # This is an initialize request
                    init_response = await self._handle_initialize(InitializeRequest(**parsed))
                    return init_response
            except Exception:
                logger.exception("Error processing JSON string")
                # Not parseable JSON, continue with normal processing
                pass

        # Check if it's a notification
        if hasattr(processed, "method"):
            method = getattr(processed, "method", None)

            # Handle notifications (methods starting with "notifications/")
            if method and method.startswith("notifications/"):
                await self._handle_notification(method, processed)
                return None

            # Handle regular methods using the dispatch table
            if method in self._method_handlers:
                # If it's a call_tool request, we need to pass the user_id
                if method == MessageMethod.CALL_TOOL:
                    return await self._method_handlers[method](processed, user_id=user_id)
                # For other methods, just pass the processed message
                return await self._method_handlers[method](processed)

            # Unknown method
            return JSONRPCError(
                id=getattr(processed, "id", None),
                error={
                    "code": -32601,
                    "message": f"Method not found: {method}",
                },
            )

        # If it's not a method request, just pass it through
        return processed

    async def _handle_notification(self, method: str, message: Any) -> None:
        """
        Handle notification messages.

        Args:
            method: The notification method
            message: The notification message
        """
        if method == "notifications/cancelled":
            logger.info(f"Request cancelled: {getattr(message, 'params', {})}")
        else:
            logger.debug(f"Received notification: {method}")

    async def _handle_ping(self, message: PingRequest) -> PingResponse:
        """
        Handle a ping request and return a pong response.

        Args:
            message: The ping request

        Returns:
            A properly formatted pong response
        """
        return PingResponse(id=message.id)

    async def _handle_initialize(self, message: InitializeRequest) -> InitializeResponse:
        """
        Handle an initialize request and return a proper initialize response.

        Args:
            message: The initialize request

        Returns:
            A properly formatted initialize response
        """
        # Create the result data
        result = InitializeResult(
            protocolVersion=MCP_PROTOCOL_VERSION,
            capabilities=ServerCapabilities(),
            serverInfo=Implementation(name="Arcade MCP Worker", version="0.1.0"),
            instructions="Arcade MCP Worker initialized.",
        )

        # Construct proper response with result field
        response = InitializeResponse(id=message.id, result=result)

        logger.debug(f"Initialize response: {response.model_dump_json()}")
        return response

    async def _handle_list_tools(
        self, message: ListToolsRequest
    ) -> Union[ListToolsResponse, JSONRPCError]:
        """
        Handle a tools/list request and return a list of available tools.

        Args:
            message: The tools/list request

        Returns:
            A properly formatted tools/list response or error
        """
        try:
            # Get all tools from the catalog
            tools = []
            tool_conversion_errors = []

            for tool in self.tool_catalog:
                try:
                    mcp_tool = create_mcp_tool(tool)
                    if mcp_tool:
                        tools.append(mcp_tool)
                except Exception:
                    tool_name = getattr(tool, "name", str(tool))
                    logger.exception(f"Error converting tool: {tool_name}")
                    tool_conversion_errors.append(tool_name)

            # Log summary if we had errors
            if tool_conversion_errors:
                logger.warning(
                    f"Failed to convert {len(tool_conversion_errors)} tools: {tool_conversion_errors}"
                )

            # Create tool objects with exception handling for each one
            tool_objects = []
            for t in tools:
                try:
                    # Make input schema optional if missing
                    tool_dict = dict(t)
                    if "inputSchema" not in tool_dict:
                        tool_dict["inputSchema"] = {"type": "object", "properties": {}}

                    tool_objects.append(Tool(**tool_dict))
                except Exception:
                    logger.exception(f"Error creating Tool object for {t.get('name', 'unknown')}")

            # Return successful response with the tools we were able to convert
            result = ListToolsResult(tools=tool_objects)
            response = ListToolsResponse(id=message.id, result=result)

        except Exception:
            logger.exception("Error listing tools")
            return JSONRPCError(
                id=message.id,
                error={
                    "code": -32603,
                    "message": "Internal error listing tools",
                },
            )
        return response

    async def _handle_call_tool(
        self, message: CallToolRequest, user_id: str | None = None
    ) -> CallToolResponse:
        """
        Handle a tools/call request to execute a tool.

        Args:
            message: The tools/call request
            user_id: Optional user ID for authentication

        Returns:
            A properly formatted tools/call response
        """
        tool_name: str = message.params["name"]
        # Extract input from the correct field
        input_params: dict[str, Any] = message.params.get("input", {})
        if not input_params:
            input_params = message.params.get("arguments", {})

        logger.info(f"Handling tool call for {tool_name}")

        try:
            tool = self.tool_catalog.get_tool_by_name(tool_name, separator="_")
            tool_context = ToolContext()

            # Set up context with secrets
            if tool.definition.requirements and tool.definition.requirements.secrets:
                self._setup_tool_secrets(tool, tool_context)

            # Handle authorization if needed
            requirement = self._get_auth_requirement(tool)
            if requirement:
                auth_result = await self._check_authorization(requirement, user_id=user_id)
                if auth_result.status != "completed":
                    return CallToolResponse(
                        id=message.id,
                        result=CallToolResult(content=[{"type": "text", "text": auth_result.url}]),
                    )
                else:
                    tool_context.authorization = ToolAuthorizationContext(
                        token=auth_result.context.token if auth_result.context else None,
                        user_info={"user_id": user_id} if user_id else {},
                    )

            # Execute the tool
            logger.debug(f"Executing tool {tool_name} with input: {input_params}")
            result = await ToolExecutor.run(
                func=tool.tool,
                definition=tool.definition,
                input_model=tool.input_model,
                output_model=tool.output_model,
                context=tool_context,
                **input_params,
            )
            logger.debug(f"Tool result: {result}")
            if result.value:
                return CallToolResponse(
                    id=message.id,
                    result=CallToolResult(content=convert_to_mcp_content(result.value)),
                )
            else:
                error = result.error or "Error calling tool"
                logger.error(f"Tool {tool_name} returned error: {error}")
                return CallToolResponse(
                    id=message.id,
                    result=CallToolResult(
                        content=[{"type": "text", "text": convert_to_mcp_content(error)}]
                    ),
                )
        except Exception as e:
            logger.exception(f"Error calling tool {tool_name}")
            error = f"Error calling tool {tool_name}: {e!s}"
            return CallToolResponse(
                id=message.id,
                result=CallToolResult(
                    content=[{"type": "text", "text": convert_to_mcp_content(error)}]
                ),
            )

    def _setup_tool_secrets(self, tool: Any, tool_context: ToolContext) -> None:
        """
        Set up tool secrets in the tool context.

        Args:
            tool: The tool to set up secrets for
            tool_context: The tool context to update
        """
        for secret in tool.definition.requirements.secrets:
            value = os.environ.get(secret.key)
            if value is not None:
                tool_context.set_secret(secret.key, value)

    async def _handle_progress(self, message: ProgressNotification) -> JSONRPCResponse:
        """
        Handle a progress notification.

        Args:
            message: The progress notification

        Returns:
            A response acknowledging the notification
        """
        return JSONRPCResponse(id=getattr(message, "id", None), result={"ok": True})

    async def _handle_cancel(self, message: CancelRequest) -> JSONRPCResponse:
        """
        Handle a cancel request.

        Args:
            message: The cancel request

        Returns:
            A response acknowledging the cancellation
        """
        return JSONRPCResponse(id=getattr(message, "id", None), result={"ok": True})

    async def _handle_shutdown(self, message: ShutdownRequest) -> ShutdownResponse:
        """
        Handle a shutdown request.

        Args:
            message: The shutdown request

        Returns:
            A response acknowledging the shutdown request
        """
        # Schedule a task to shutdown the server after sending the response
        proc = asyncio.create_task(self.shutdown())
        proc.add_done_callback(lambda _: logger.info("MCP server shutdown complete"))
        return ShutdownResponse(id=message.id, result={"ok": True})

    async def _handle_list_resources(self, message: ListResourcesRequest) -> ListResourcesResponse:
        """
        Handle a resources/list request.

        Args:
            message: The resources/list request

        Returns:
            A properly formatted resources/list response
        """
        return ListResourcesResponse(id=message.id, result={"resources": []})

    async def _handle_list_prompts(self, message: ListPromptsRequest) -> ListPromptsResponse:
        """
        Handle a prompts/list request.

        Args:
            message: The prompts/list request

        Returns:
            A properly formatted prompts/list response
        """
        return ListPromptsResponse(id=message.id, result={"prompts": []})

    def _get_auth_requirement(self, tool: MaterializedTool) -> AuthRequirement | None:
        """
        Get the authentication requirement for a tool.

        Args:
            tool: The tool to get the requirement for

        Returns:
            An authentication requirement or None if not required
        """
        req = tool.definition.requirements.authorization
        if not req:
            return None
        if not req.provider_id and not req.provider_type:
            return None
        if hasattr(req, "oauth2") and req.oauth2:
            return AuthRequirement(
                provider_id=str(req.provider_id),
                provider_type=str(req.provider_type),
                oauth2=AuthRequirementOauth2(scopes=req.oauth2.scopes or []),
            )
        return AuthRequirement(
            provider_id=str(req.provider_id),
            provider_type=str(req.provider_type),
        )

    async def _check_authorization(
        self, auth_requirement: AuthRequirement, user_id: str | None = None
    ) -> AuthorizationResponse:
        """
        Check if a tool is authorized for a user.

        Args:
            tool: The tool to check authorization for
            user_id: The user ID to check authorization for

        Returns:
            An authorization response

        Raises:
            RuntimeError: If the tool has no authorization requirement
            Exception: If authorization fails
        """
        try:
            response = await self.arcade.auth.authorize(
                auth_requirement=auth_requirement,
                user_id=user_id or "anonymous",
            )
            logger.debug(f"Authorization response: {response}")

        except ArcadeError:
            logger.exception("Error authorizing tool")
            raise
        return response

    async def shutdown(self) -> None:
        """Shutdown the server."""
        self._shutdown = True
        logger.info("MCP server shutdown complete")



import json

from arcade_evals import (
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_evals.critic import BinaryCritic
from arcade_tdk import ToolCatalog

import arcade_jira
from arcade_jira.tools.issues import (
    get_issue_type_by_id,
    list_issue_types_by_project,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_jira)


@tool_eval()
def list_issue_types_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="List issue types eval suite",
        system_message=(
            "You are an AI assistant with access to Jira tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="List issue types in a project (ID)",
        user_message="List the issue types in the project with ID '1234567890'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_issue_types_by_project,
                args={
                    "project": "1234567890",
                    "limit": 200,
                    "offset": 0,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="project", weight=0.8),
            BinaryCritic(critic_field="limit", weight=0.1),
            BinaryCritic(critic_field="offset", weight=0.1),
        ],
    )

    suite.add_case(
        name="List issue types in a project (Key)",
        user_message="List the issue types in the project PRJ-1.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_issue_types_by_project,
                args={
                    "project": "PRJ-1",
                    "limit": 200,
                    "offset": 0,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="project", weight=0.8),
            BinaryCritic(critic_field="limit", weight=0.1),
            BinaryCritic(critic_field="offset", weight=0.1),
        ],
    )

    suite.add_case(
        name="List issue types in a project (name)",
        user_message="List the issue types in the project 'Engineering'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_issue_types_by_project,
                args={
                    "project": "Engineering",
                    "limit": 200,
                    "offset": 0,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="project", weight=0.8),
            BinaryCritic(critic_field="limit", weight=0.1),
            BinaryCritic(critic_field="offset", weight=0.1),
        ],
    )

    suite.add_case(
        name="List issue types in a project (name) with pagination in the prompt",
        user_message="List 10 issue types in the project 'Engineering'. Skip the first 30 items.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_issue_types_by_project,
                args={
                    "project": "Engineering",
                    "limit": 10,
                    "offset": 30,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="project", weight=1 / 3),
            BinaryCritic(critic_field="limit", weight=1 / 3),
            BinaryCritic(critic_field="offset", weight=1 / 3),
        ],
    )

    suite.add_case(
        name="List issue types in a project (name) with pagination from chat history",
        user_message="Get the next items.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_issue_types_by_project,
                args={
                    "project": "Engineering",
                    "limit": 2,
                    "offset": 2,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="project", weight=1 / 3),
            BinaryCritic(critic_field="limit", weight=1 / 3),
            BinaryCritic(critic_field="offset", weight=1 / 3),
        ],
        additional_messages=[
            {
                "role": "user",
                "content": "List 2 issue types in the project 'Engineering'.",
            },
            {
                "role": "assistant",
                "content": "",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "Jira_ListIssueTypesByProject",
                            "arguments": json.dumps({
                                "project": "Engineering",
                                "limit": 2,
                                "offset": 0,
                            }),
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "content": json.dumps({
                    "project": {
                        "id": "10000",
                        "key": "Eng-1",
                        "name": "Engineering",
                    },
                    "issue_types": [
                        {
                            "id": "10001",
                            "name": "Bug",
                            "description": (
                                "A bug is an error or flaw in a software application or website."
                            ),
                        },
                        {
                            "id": "10002",
                            "name": "Task",
                            "description": "A task is a unit of work that needs to be completed.",
                        },
                    ],
                    "pagination": {
                        "limit": 2,
                        "total_results": 2,
                        "next_offset": 2,
                    },
                }),
                "tool_call_id": "call_1",
                "name": "Jira_ListIssueTypesByProject",
            },
            {
                "role": "assistant",
                "content": (
                    "Here are two issue types in the project 'Engineering':\n\n1. Bug\n2. Task"
                ),
            },
        ],
    )

    return suite


@tool_eval()
def get_issue_type_by_id_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Get issue type by ID eval suite",
        system_message=(
            "You are an AI assistant with access to Jira tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get issue type by ID",
        user_message="Get the issue type with ID '10001'.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_issue_type_by_id,
                args={
                    "issue_type": "10001",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="project", weight=0.8),
            BinaryCritic(critic_field="limit", weight=0.1),
            BinaryCritic(critic_field="offset", weight=0.1),
        ],
    )

    return suite



from enum import Enum


# Models and enums for the e2b code interpreter
class E2BSupportedLanguage(str, Enum):
    PYTHON = "python"
    JAVASCRIPT = "js"
    R = "r"
    JAVA = "java"
    BASH = "bash"
